__all__ = [
	'WavefrontSensorOptics',
	'WavefrontSensorEstimator',
	'make_odwfs_amplitude_filter',
	'make_polarization_odwfs_amplitude_filter',
	'optical_differentiation_surface',
	'OpticalDifferentiationWavefrontSensorOptics',
	'OpticalDifferentiationWavefrontSensorEstimator',
	'ModulatedPyramidWavefrontSensorOptics',
	'PyramidWavefrontSensorOptics',
	'PyramidWavefrontSensorEstimator',
	'ShackHartmannWavefrontSensorOptics',
	'SquareShackHartmannWavefrontSensorOptics',
	'ShackHartmannWavefrontSensorEstimator',
	'ZernikeWavefrontSensorOptics',
	'ZernikeWavefrontSensorEstimator',
	'VectorZernikeWavefrontSensorOptics'
]

from .optical_differentiation_wavefront_sensor import *
from .pyramid import *
from .shack_hartmann import *
from .wavefront_sensor import *
from .zernike_wavefront_sensor import *



__all__ = [
	'get_strehl_from_focal',
	'get_strehl_from_pupil',
	'get_mean_intensity_in_roi',
	'get_mean_raw_contrast',
	'binned_profile',
	'azimutal_profile',
	'radial_profile'
]

from .contrast import *
from .profile import *



__all__ = [
	'GifWriter',
	'FFMpegWriter',
	'set_color_scheme',
	'colors',
	'errorfill',
	'imshow_field',
	'imsave_field',
	'contour_field',
	'contourf_field',
	'complex_field_to_rgb',
	'imshow_psf',
	'imshow_pupil_phase'
]

from .animation import *
from .color_scheme import *
from . import colors
from .error_bars import *
from .field import *
from .util import *



import numpy as np

from .wavefront import Wavefront, jones_to_mueller
from .optical_element import OpticalElement, make_agnostic_forward, make_agnostic_backward, AgnosticOpticalElement
from ..field import Field, field_dot, field_conjugate_transpose

def rotation_matrix(angle):
	'''Two dimensional rotation matrix.

	Parameters
	----------
	angle : scaler
		rotation angle in radians

	Returns
	-------
	ndarray
		The rotation matrix.
	'''
	return np.array([[np.cos(angle),  np.sin(angle)],[-np.sin(angle), np.cos(angle)]])

class JonesMatrixOpticalElement(AgnosticOpticalElement):
	'''A general Jones Matrix.

	Parameters
	----------
	jones_matrix : matrix or Field
		The Jones matrix describing the optical element.
	'''
	def __init__(self, jones_matrix):
		self.jones_matrix = jones_matrix

		AgnosticOpticalElement.__init__(self, True, True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		instance_data.jones_matrix = self.evaluate_parameter(self.jones_matrix, input_grid, output_grid, wavelength)

	def get_input_grid(self, output_grid, wavelength):
		return output_grid

	def get_output_grid(self, input_grid, wavelength):
		return input_grid

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		'''Propgate the wavefront through the Jones matrix.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The propagated wavefront.
		'''
		if wavefront.is_scalar:
			# We generate an unpolarized wavefront
			if hasattr(instance_data.jones_matrix, 'grid'):
				electric_field = instance_data.jones_matrix * wavefront.electric_field
			else:
				electric_field = instance_data.jones_matrix[..., np.newaxis] * wavefront.electric_field

			return Wavefront(electric_field, wavelength=wavefront.wavelength, input_stokes_vector=[1, 0, 0, 0])
		else:
			wf = wavefront.copy()
			wf.electric_field = field_dot(instance_data.jones_matrix, wf.electric_field)

			return wf

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		'''Propagate the wavefront backwards through the Jones matrix.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The propagated wavefront.
		'''
		if wavefront.is_scalar:
			# We generate an unpolarized wavefront
			if hasattr(instance_data.jones_matrix, 'grid'):
				electric_field = field_conjugate_transpose(instance_data.jones_matrix) * wavefront.electric_field
			else:
				electric_field = field_conjugate_transpose(instance_data.jones_matrix)[..., np.newaxis] * wavefront.electric_field

			return Wavefront(electric_field, wavelength=wavefront.wavelength, input_stokes_vector=[1, 0, 0, 0])
		else:
			wf = wavefront.copy()
			wf.electric_field = field_dot(field_conjugate_transpose(instance_data.jones_matrix), wf.electric_field)

			return wf

	@property
	def mueller_matrix(self):
		'''Returns the Mueller matrix corresponding to the Jones matrix.
		'''
		return self.construct_function(jones_to_mueller, self.jones_matrix)

	def __mul__(self, b):
		'''Multiply the Jones matrix on the right side with the Jones matrix m.

		Parameters
		----------
		m : JonesMatrixOpticalElement, Field or matrix
			The Jones matrix with which to multiply.
		'''
		if hasattr(b, 'jones_matrix'):
			jones_matrix = self.construct_function(field_dot, self.jones_matrix, b.jones_matrix)
		else:
			jones_matrix = self.construct_function(field_dot, self.jones_matrix, b)

		return JonesMatrixOpticalElement(jones_matrix)

class PhaseRetarder(JonesMatrixOpticalElement):
	'''A general phase retarder.

	Parameters
	----------
	phase_retardation : scalar or Field
		The relative phase retardation induced between the fast and slow axis.
	fast_axis_orientation : scalar or Field
		The angle of the fast axis with respect to the x-axis in radians.
	circularity : scalar or Field
		The circularity of the phase retarder.
	'''
	def __init__(self, phase_retardation, fast_axis_orientation, circularity):
		self._phase_retardation = phase_retardation
		self._fast_axis_orientation = fast_axis_orientation
		self._circularity = circularity

		JonesMatrixOpticalElement.__init__(self, self.jones_matrix)

	@property
	def jones_matrix(self):
		def jones(phase_retardation, fast_axis_orientation, circularity):
			phi_plus = np.exp(1j * phase_retardation / 2)
			phi_minus = np.exp(-1j * phase_retardation / 2)

			# calculating the individual components
			j11 = phi_plus * np.cos(fast_axis_orientation)**2 + phi_minus * np.sin(fast_axis_orientation)**2
			j12 = (phi_plus - phi_minus) * np.exp(-1j * circularity) * np.cos(fast_axis_orientation) * np.sin(fast_axis_orientation)
			j21 = (phi_plus - phi_minus) * np.exp(1j * circularity) * np.cos(fast_axis_orientation) * np.sin(fast_axis_orientation)
			j22 = phi_plus * np.sin(fast_axis_orientation)**2 + phi_minus * np.cos(fast_axis_orientation)**2

			# constructing the Jones matrix.
			jones_matrix = np.array([[j11, j12], [j21, j22]])

			if hasattr(j11, 'grid'):
				jones_matrix = Field(jones_matrix, j11.grid)

			return jones_matrix

		return self.construct_function(jones, self.phase_retardation, self.fast_axis_orientation, self.circularity)

	@jones_matrix.setter
	def jones_matrix(self, b):
		pass

	@property
	def phase_retardation(self):
		return self._phase_retardation

	@phase_retardation.setter
	def phase_retardation(self, phase_retardation):
		self._phase_retardation = phase_retardation

		self.clear_cache()

	@property
	def fast_axis_orientation(self):
		return self._fast_axis_orientation

	@fast_axis_orientation.setter
	def fast_axis_orientation(self, fast_axis_orientation):
		self._fast_axis_orientation = fast_axis_orientation

		self.clear_cache()

	@property
	def circularity(self):
		return self._circularity

	@circularity.setter
	def circularity(self, circularity):
		self._circularity = circularity

		self.clear_cache()

class LinearRetarder(PhaseRetarder):
	'''A general linear retarder.

	Parameters
	----------
	phase_retardation : scalar or Field
		The relative phase retardation induced between the fast and slow axis.
	fast_axis_orientation : scalar or Field
		The angle of the fast axis with respect to the x-axis in radians.
	'''
	def __init__(self, phase_retardation, fast_axis_orientation):
		PhaseRetarder.__init__(self, phase_retardation, fast_axis_orientation, 0)

class CircularRetarder(PhaseRetarder):
	'''A general circular retarder.

	Parameters
	----------
	phase_retardation : scalar or Field
		The relative phase retardation induced between the fast and slow axis.
	fast_axis_orientation : scalar or Field
		The angle of the fast axis with respect to the x-axis in radians.
	'''
	def __init__(self, phase_retardation):
		PhaseRetarder.__init__(self, phase_retardation, np.pi / 4, np.pi / 2)

class QuarterWavePlate(LinearRetarder):
	'''A quarter-wave plate.

	Parameters
	----------
	fast_axis_orientation : scalar or Field
		The angle of the fast axis with respect to the x-axis in radians.
	'''
	def __init__(self, fast_axis_orientation):
		LinearRetarder.__init__(self, np.pi / 2, fast_axis_orientation)

class HalfWavePlate(LinearRetarder):
	'''A half-wave plate.

	Parameters
	----------
	fast_axis_orientation : scalar or Field
		The angle of the fast axis with respect to the x-axis in radians.
	'''
	def __init__(self, fast_axis_orientation):
		LinearRetarder.__init__(self, np.pi, fast_axis_orientation)

class GeometricPhaseElement(LinearRetarder):
	'''A general geometric phase element.

	Parameters
	----------
	phase_pattern : Field or array_like
		The phase pattern in radians.
	leakage : scalar
		The relative leakage strength (0 = no leakage, 1 = maximum leakage)
	retardance_offset : scalar
		The retardance offset from half wave in radians. This will result in leakage.
	'''
	def __init__(self, phase_pattern, leakage=None, retardance_offset=0):
		if leakage is not None:
			# Testing if the leakage is valid.
			if leakage < 0 or leakage > 1:
				raise ValueError('Leakage must be between 0 and 1.')

			# Calculating the required retardance offset to get the leakage.
			retardance_offset = 2 * np.arcsin(np.sqrt(leakage))

		LinearRetarder.__init__(self, np.pi - retardance_offset, phase_pattern / 2)

class LinearPolarizer(JonesMatrixOpticalElement):
	'''A linear polarizer.

	Parameters
	----------
	polarization_angle : scalar or Field
		The polarization angle of the polarizer. Light with this angle is transmitted.
	'''
	def __init__(self, polarization_angle):
		self.polarization_angle = polarization_angle

		JonesMatrixOpticalElement.__init__(self, self.jones_matrix)

	@property
	def jones_matrix(self):
		def jones(angle):
			c = np.cos(angle)
			s = np.sin(angle)

			return np.array([[c**2, c * s],[c * s, s**2]])

		return self.construct_function(jones, self.polarization_angle)

	@jones_matrix.setter
	def jones_matrix(self, b):
		pass

	@property
	def polarization_angle(self):
		'''The angle of polarization of the linear polarizer.
		'''
		return self._polarization_angle

	@polarization_angle.setter
	def polarization_angle(self, angle):
		self._polarization_angle = angle

		self.clear_cache()

class LinearPolarizingBeamSplitter(OpticalElement):
	''' A linear polarizing beam splitter that accepts one wavefront and returns two orthogonally linearly polarized wavefronts.

	The first of the returned wavefront will have the polarization state set by the polarization angle. The second wavefront
	will be perpendicular to the first.

	Parameters
	----------
	polarization_angle : scalar or Field
		The polarization angle of the polarizer.
	'''
	def __init__(self, polarization_angle, wavelength=1):
		self.polarization_angle = polarization_angle

	@property
	def polarization_angle(self):
		'''The angle of polarization of the linear polarizer.
		'''
		return self._polarization_angle

	@polarization_angle.setter
	def polarization_angle(self, polarization_angle):
		self._polarization_angle = polarization_angle

		# Calculating two Jones matrices for the two ports.
		self.polarizer_port_1 = LinearPolarizer(polarization_angle)
		self.polarizer_port_2 = LinearPolarizer(polarization_angle + np.pi / 2)

	def forward(self, wavefront):
		'''Propgate the wavefront through the PBS.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		wf_1 : Wavefront
			The wavefront propagated through a polarizer under the polarization angle.

		wf_2 : Wavefront
			The propagated wavefront through a polarizer perpendicular to the first one.
		'''

		wf_1 = self.polarizer_port_1.forward(wavefront)
		wf_2 = self.polarizer_port_2.forward(wavefront)

		return wf_1, wf_2

	def backward(self, wavefront):
		'''Propagate the wavefront backwards through the PBS.

		Not possible, will raise error.
		'''
		raise RuntimeError('Backward propagation through PolarizingBeamSplitter not possible.')

	@property
	def mueller_matrices(self):
		'''Returns the Mueller matrices of the two Jones matrices.
		'''
		return jones_to_mueller(self.polarizer_port_1.jones_matrix), jones_to_mueller(self.polarizer_port_2.jones_matrix)

class CircularPolarizingBeamSplitter(OpticalElement):
	''' A circular polarizing beam splitter that accepts one wavefront and returns two linearly polarized wavefronts.

	The circular polarizing beam splitter is a combination of a quarter-wave plate and a polarizing beam splitter. The
	quarter-wave plate rotation angle is 45 degrees and the polarizing beam-splitter rotation angle is zero degrees.
	Therefore, the first of the returned wavefronts will have a linear polarization state with an electric field equal
	to the amount of left-circular polarization in the incoming wavefront. The second wavefront will be perpendicular
	to the first and with electric field equal to the amount of right-circular polarization in the incoming wavefront.

	Parameters
	----------
	polarization_angle : scalar or Field
		The polarization angle of the polarizer.
	'''
	def __init__(self, wavelength=1):
		self.polarization_angle = 0
		self.quarter_wave_angle = np.pi / 4
		self.quarter_wave_plate = QuarterWavePlate(self.quarter_wave_angle)
		self.linear_polarizing_beam_splitter = LinearPolarizingBeamSplitter(self.polarization_angle)

	def forward(self, wavefront):
		'''Propgate the wavefront through the CBS.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		wf_1 : Wavefront
			The wavefront propagated through a quarter-wave plate at 45 degrees and a polarizer 0 degrees.
		wf_2 : Wavefront
			The wavefront propagated through a quarter-wave plate at 45 degrees and a polarizer 90 degrees.
		'''
		wf_1, wf_2 = self.linear_polarizing_beam_splitter.forward(self.quarter_wave_plate.forward(wavefront))
		return wf_1, wf_2

	def backward(self, wavefront):
		'''Propagate the wavefront backwards through the CBS.

		Not possible, will raise error.
		'''
		raise RuntimeError('Backward propagation through PolarizingBeamSplitter not possible.')

	@property
	def mueller_matrices(self):
		'''Returns the Mueller matrices of the two Jones matrices.
		'''
		qwp_mueller = self.quarter_wave_plate.mueller_matrix
		lin_pol_bs_mueller_1 , lin_pol_bs_mueller_2 = self.linear_polarizing_beam_splitter.mueller_matrices

		return np.dot(lin_pol_bs_mueller_1, qwp_mueller), np.dot(lin_pol_bs_mueller_2, qwp_mueller)



from hcipy import *
from matplotlib import pyplot as plt
import numpy as np

def test_pupil_plane_wavefront_sensors():
	# Create the input grid
	D = 1
	pupil_grid = make_pupil_grid(128, D)
	aperture = circular_aperture(D)

	# Make zernike basis
	num_zernike = 3
	zernike_basis = make_zernike_basis(num_zernike, D, pupil_grid, 2)

	# Create wavefront sensors
	pyramid_wfs = PyramidWavefrontSensorOptics(pupil_grid)
	zernike_wfs = ZernikeWavefrontSensorOptics(pupil_grid, q=32, num_airy=20)
	god_wfs = gODWavefrontSensorOptics(40, 1/3, pupil_grid, pupil_separation=1.5, q=16, num_airy=40)
	pod_wfs = PolgODWavefrontSensorOptics(40, 1/3, pupil_grid, pupil_separation=1.5, q=16, num_airy=40)

	wavefront_sensors = (pyramid_wfs, zernike_wfs, god_wfs, pod_wfs)
	num_wfs = len(wavefront_sensors)

	# Create reference images
	wf = Wavefront(aperture(pupil_grid))
	wf.total_power = 1
	refs = [wfs.forward(wf).intensity for wfs in wavefront_sensors]
	eff = [wfs.forward(wf).total_power for wfs in wavefront_sensors]
	print("Total efficiency : ", eff)

	for i in range(num_wfs):
		plt.subplot(2,2,i+1)
		imshow_field(refs[i], wavefront_sensors[i].output_grid)
	plt.show()

	for i, zern in enumerate(zernike_basis):
		# create input
		wf = Wavefront(aperture(pupil_grid) * np.exp(1j * 0.01 * zern))
		wf.total_power = 1

		# Measure the response of all wavefront sensors
		for j, wfs in enumerate(wavefront_sensors):
			wf_out = wfs.forward(wf)
			print(wf_out.total_power)
			plt.subplot(num_wfs, num_zernike, i+1 + num_zernike*j)
			imshow_field(wf_out.intensity-refs[j])

	plt.show()



if __name__ == "__main__":
	test_pupil_plane_wavefront_sensors()



import heapq

class DynamicOpticalSystem(object):
	'''A dynamically varying optical system.

	This can be used as a starting point for an implementation of an adaptive optics system. The optics
	are changed as specific moments in time, which can be scheduled by submitting a callback. The time
	points are bridged using the integrate function, which is written by the derived class. This
	function should integrate on each detector without reading out the detector. Detector readouts or
	DM changes should be implemented as callbacks.
	'''
	def __init__(self):
		self.callbacks = []
		self.t = 0
		self.callback_counter = 0

	def evolve_until(self, t):
		'''Evolve the optical system until time `t`.

		Parameters
		----------
		t : scalar
			The point in time to which to simulate this optical system. Callbacks and integrations
			are handled along the way.
		'''
		if t < self.t:
			raise ValueError('Backwards evolution is not allowed.')

		end = False
		while not end:
			t_next, _, callback = self.callbacks[0]
			if t_next < t:
				integration_time = t_next - self.t
				heapq.heappop(self.callbacks)
			else:
				integration_time = t - self.t
				end = True

			# Avoid multiple expensive integrations if we have multiple callbacks
			# at the same time.
			if integration_time > 1e-6:
				self.integrate(integration_time)
				self.t += integration_time

			if not end:
				callback()

	def integrate(self, integration_time):
		'''Integrate the current optical system for a certain integration time.

		This function should be implemented by a user in a derived class.

		Parameters
		----------
		integration_time : scalar
			The integration time with which to integrate the optical system.
		'''
		pass

	def add_callback(self, t, callback):
		'''Add a callback to the list of callbacks.

		This function can even be called during handling of a callback. This is
		especially useful when implementing periodic callbacks: a callback can
		reinsert itself at a later time at the end of handling of the callback.

		Parameters
		----------
		t : scalar
			The time at which to call the callback.
		callback : function
			The function to call at time `t`. This function should have no arguments.
		'''
		# Callback counter is to avoid comparison of callback functions
		# if the times are equal, as comparison of functions is not allowed.
		heapq.heappush(self.callbacks, (t, self.callback_counter, callback))
		self.callback_counter += 1



from scipy.interpolate import RegularGridInterpolator, NearestNDInterpolator
import numpy as np
from ..field import Field

def make_nearest_interpolator_separated(field, grid=None):
	'''Make a nearest interpolator for a field on a separated grid.

	Parameters
	----------
	field : Field or ndarray
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.

	Returns
	-------
	Field generator
		The interpolator, as a Field generator. The grid on which this field generator will evaluated, does
		not have to be separated.
	'''
	if grid is None:
		grid = field.grid
	else:
		field = Field(field, grid)

	axes_reversed = np.array(grid.separated_coords)
	interp = RegularGridInterpolator(axes_reversed, field.shaped, 'nearest', False)

	def interpolator(evaluated_grid):
		evaluated_coords = np.flip(np.array(evaluated_grid.coords), 0)
		res = interp(evaluated_coords.T)
		return Field(res.ravel(), evaluated_grid)

	return interpolator

def make_nearest_interpolator_unstructured(field, grid=None):
	'''Make a nearest interpolator for an unstructured grid.

	Parameters
	----------
	field : Field or array_like
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.

	Returns
	-------
	Field generator
		The interpolator as a Field generator. The grid on which this field generator will be evaluated does
		not need to have any structure.
	'''
	if grid is None:
		grid = field.grid
	else:
		field = Field(field, grid)

	interp = NearestNDInterpolator(grid.points, field)

	def interpolator(evaluated_grid):
		res = interp(grid.points)
		return Field(res, evaluated_grid)

	return interpolator

def make_nearest_interpolator(field, grid=None):
	'''Make a nearest interpolator for any type of grid.

	Parameters
	----------
	field : Field or array_like
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.

	Returns
	-------
	Field generator
		The interpolator as a Field generator. The grid on which this field generator will be evaluated does
		not need to have any structure.
	'''
	if grid is None:
		grid = field.grid

	if grid.is_unstructured:
		return make_nearest_interpolator_unstructured(field, grid)
	else:
		return make_nearest_interpolator_separated(field, grid)



from __future__ import division

from .atmospheric_model import AtmosphericLayer, phase_covariance_von_karman, fried_parameter_from_Cn_squared
from ..field import Field, RegularCoords, UnstructuredCoords, CartesianGrid
from .finite_atmospheric_layer import FiniteAtmosphericLayer

import numpy as np
from scipy import linalg
from scipy.ndimage import affine_transform

import warnings

class InfiniteAtmosphericLayer(AtmosphericLayer):
	def __init__(self, input_grid, Cn_squared=None, L0=np.inf, velocity=0, height=0, stencil_length=2, use_interpolation=True):
		self._initialized = False

		AtmosphericLayer.__init__(self, input_grid, Cn_squared, L0, velocity, height)

		# Check properties of input_grid
		if not input_grid.is_('cartesian'):
			raise ValueError('Input grid must be cartesian.')
		if not input_grid.is_regular:
			raise ValueError('Input grid must be regularly spaced')
		if not input_grid.ndim == 2:
			raise ValueError('Input grid must be two-dimensional.')

		self.stencil_length = stencil_length
		self.use_interpolation = use_interpolation

		self._make_stencils()
		self._make_covariance_matrices()
		self._make_AB_matrices()
		self._make_initial_phase_screen()

		self.center = np.zeros(2)
		self._t = 0

		self._initialized = True

	def _recalculate_matrices(self):
		if self._initialized:
			self._make_covariance_matrices()
			self._make_AB_matrices()

	def _make_stencils(self):
		# Vertical
		zero = self.input_grid.zero - np.array([0, self.input_grid.delta[1]])
		self.new_grid_bottom = CartesianGrid(RegularCoords(self.input_grid.delta, [self.input_grid.dims[0], 1], zero))

		self.stencil_bottom = Field(np.zeros(self.input_grid.size, dtype='bool'), self.input_grid).shaped
		self.stencil_bottom[:self.stencil_length,:] = True

		for i, n in enumerate(np.random.geometric(0.5, self.input_grid.dims[0])):
			self.stencil_bottom[(n + self.stencil_length - 1) % self.input_grid.dims[1],i] = True

		self.stencil_bottom = self.stencil_bottom.ravel()
		self.num_stencils_vertical = np.sum(self.stencil_bottom)

		# Horizontal
		zero = self.input_grid.zero - np.array([self.input_grid.delta[0], 0])
		self.new_grid_left = CartesianGrid(RegularCoords(self.input_grid.delta, [1, self.input_grid.dims[1]], zero))

		self.stencil_left = Field(np.zeros(self.input_grid.size, dtype='bool'), self.input_grid).shaped
		self.stencil_left[:,:self.stencil_length] = True

		for i, n in enumerate(np.random.geometric(0.5, self.input_grid.dims[1])):
			self.stencil_left[i,(n + self.stencil_length - 1) % self.input_grid.dims[0]] = True

		self.stencil_left = self.stencil_left.ravel()
		self.num_stencils_horizontal = np.sum(self.stencil_left)

	def _make_covariance_matrices(self):
		phase_covariance = phase_covariance_von_karman(fried_parameter_from_Cn_squared(1, 1), self.L0)

		# Vertical
		x = np.concatenate((self.input_grid.x[self.stencil_bottom], self.new_grid_bottom.x))
		x = np.concatenate([x - xx for xx in x])
		y = np.concatenate((self.input_grid.y[self.stencil_bottom], self.new_grid_bottom.y))
		y = np.concatenate([y - yy for yy in y])

		separations = CartesianGrid(UnstructuredCoords((x, y)))
		n = self.new_grid_bottom.size + self.num_stencils_vertical
		self.cov_matrix_vertical = phase_covariance(separations).reshape((n, n))

		# Horizontal
		x = np.concatenate((self.input_grid.x[self.stencil_left], self.new_grid_left.x))
		x = np.concatenate([x - xx for xx in x])
		y = np.concatenate((self.input_grid.y[self.stencil_left], self.new_grid_left.y))
		y = np.concatenate([y - yy for yy in y])

		separations = CartesianGrid(UnstructuredCoords((x, y)))
		n = self.new_grid_left.size + self.num_stencils_horizontal
		self.cov_matrix_horizontal = phase_covariance(separations).reshape((n, n))

	def _make_AB_matrices(self):
		# Vertical
		n = self.num_stencils_vertical
		cov_zz = self.cov_matrix_vertical[:n,:n]
		cov_xz = self.cov_matrix_vertical[n:, :n]
		cov_zx = self.cov_matrix_vertical[:n, n:]
		cov_xx = self.cov_matrix_vertical[n:, n:]

		cf = linalg.cho_factor(cov_zz)
		inv_cov_zz = linalg.cho_solve(cf, np.eye(cov_zz.shape[0]))

		self.A_vertical = cov_xz.dot(inv_cov_zz)

		BBt = cov_xx - self.A_vertical.dot(cov_zx)

		U, S, Vt = np.linalg.svd(BBt)
		L = np.sqrt(S[:self.input_grid.dims[0]])

		self.B_vertical = U * L

		# Horizontal
		n = self.num_stencils_horizontal
		cov_zz = self.cov_matrix_horizontal[:n,:n]
		cov_xz = self.cov_matrix_horizontal[n:, :n]
		cov_zx = self.cov_matrix_horizontal[:n, n:]
		cov_xx = self.cov_matrix_horizontal[n:, n:]

		cf = linalg.cho_factor(cov_zz)
		inv_cov_zz = linalg.cho_solve(cf, np.eye(cov_zz.shape[0]))

		self.A_horizontal = cov_xz.dot(inv_cov_zz)

		BBt = cov_xx - self.A_horizontal.dot(cov_zx)

		U, S, Vt = np.linalg.svd(BBt)
		L = np.sqrt(S[:self.input_grid.dims[1]])

		self.B_horizontal = U * L

	def _make_initial_phase_screen(self):
		oversampling = 16
		layer = FiniteAtmosphericLayer(self.input_grid, self.Cn_squared, self.outer_scale, self.velocity, self.height, oversampling)
		self._achromatic_screen = layer.phase_for(1)
		self._shifted_achromatic_screen = self._achromatic_screen

	def _extrude(self, where=None):
		flipped = (where == 'top') or (where == 'right')
		horizontal = (where == 'left') or (where == 'right')

		if where == 'top' or where == 'right':
			screen = self._achromatic_screen[::-1]
		else:
			screen = self._achromatic_screen

		if horizontal:
			stencil = self.stencil_left
			A = self.A_horizontal
			B = self.B_horizontal
		else:
			stencil = self.stencil_bottom
			A = self.A_vertical
			B = self.B_vertical

		stencil_data = screen[stencil]
		random_data = np.random.normal(0, 1, size=B.shape[1])
		new_slice = A.dot(stencil_data) + B.dot(random_data) * np.sqrt(self._Cn_squared)

		screen = screen.shaped

		if horizontal:
			screen = np.hstack((new_slice[:,np.newaxis], screen[:,:-1]))
		else:
			screen = np.vstack((new_slice[np.newaxis,:], screen[:-1,:]))

		screen = Field(screen, self.input_grid)

		if flipped:
			self._achromatic_screen = screen[::-1,::-1].ravel()
		else:
			self._achromatic_screen = screen.ravel()

	def phase_for(self, wavelength):
		return self._shifted_achromatic_screen / wavelength

	def reset(self):
		self._make_initial_phase_screen()
		self.center = np.zeros(2)
		self._t = 0

	def evolve_until(self, t):
		if t is None:
			self.reset()
			return

		# Get the old and new center positions
		old_center = np.round(self.center / self.input_grid.delta).astype('int')
		self.center = self.center + self.velocity * (t - self._t)
		new_pixel_center = np.round(self.center / self.input_grid.delta).astype('int')

		self._t = t

		# Measure the number of full pixel shifts
		delta = new_pixel_center - old_center
		# Measure the sub-pixel shift
		sub_delta = self.center - new_pixel_center * self.input_grid.delta

		for i in range(abs(delta[0])):
			if delta[0] < 0:
				self._extrude('left')
			else:
				self._extrude('right')

		for i in range(abs(delta[1])):
			if delta[1] < 0:
				self._extrude('bottom')
			else:
				self._extrude('top')

		if self.use_interpolation:
			# Use bilinear interpolation to interpolate the achromatic phase screen to the correct position.
			# This is to avoid sudden shifts by discrete pixels.
			ps = self._achromatic_screen.shaped
			with warnings.catch_warnings():
				# Suppress warnings about the changed behaviour in affine_transform for 1D arrays.
				# We know about this and are expecting the behaviour as is.
				warnings.filterwarnings('ignore', message='The behaviour of affine_transform')
				warnings.filterwarnings('ignore', message='The behavior of affine_transform')

				screen = affine_transform(ps, np.array([1,1]), (sub_delta / self.input_grid.delta)[::-1], mode='nearest', order=5)
				self._shifted_achromatic_screen = Field(screen.ravel(), self._achromatic_screen.grid)
		else:
			self._shifted_achromatic_screen = self._achromatic_screen

	@property
	def Cn_squared(self):
		return self._Cn_squared

	@Cn_squared.setter
	def Cn_squared(self, Cn_squared):
		self._Cn_squared = Cn_squared

	@property
	def outer_scale(self):
		return self._L0

	@outer_scale.setter
	def L0(self, L0):
		self._L0 = L0

		self._recalculate_matrices()



import numpy as np
import copy

class Coords(object):
	'''Base class for coordinates.
	'''
	_coordinate_types = {}

	def copy(self):
		'''Make a copy.
		'''
		return copy.deepcopy(self)

	@classmethod
	def from_dict(cls, tree):
		coordinate_class = Coords._coordinate_types[tree['type']]

		return coordinate_class.from_dict(tree)

	def to_dict(self):
		raise NotImplementedError()

	def __add__(self, b):
		'''Add `b` to the coordinates separately and return the result.
		'''
		res = self.copy()
		res += b
		return res

	def __iadd__(self, b):
		'''Add `b` to the coordinates separately in-place.
		'''
		raise NotImplementedError()

	def __radd__(self, b):
		'''Add `b` to the coordinates separately and return the result.
		'''
		return self + b

	def __sub__(self, b):
		'''Subtract `b` from the coordinates separately and return the result.
		'''
		return self + (-b)

	def __isub__(self, b):
		'''Subtract `b` from the coordinates separately in-place.
		'''
		self += (-b)
		return self

	def __mul__(self, f):
		'''Multiply each coordinate with `f` separately and return the result.
		'''
		res = self.copy()
		res *= f
		return res

	def __rmul__(self, f):
		'''Multiply each coordinate with `f` separately and return the result.
		'''
		return self * f

	def __imul__(self, f):
		'''Multiply each coordinate with `f` separately in-place.
		'''
		raise NotImplementedError()

	def __div__(self, f):
		'''Divide each coordinate with `f` separately and return the result.
		'''
		return self * (1./f)

	def __idiv__(self, f):
		'''Divide each coordinate with `f` separately in-place.
		'''
		self *= (1./f)
		return self

	def __getitem__(self, i):
		'''The `i`-th point for these coordinates.
		'''
		raise NotImplementedError()

	@property
	def is_separated(self):
		'''True if the coordinates are separated, False otherwise.
		'''
		return hasattr(self, 'separated_coords')

	@property
	def is_regular(self):
		'''True if the coordinates are regularly-spaced, False otherwise.
		'''
		return hasattr(self, 'regular_coords')

	@property
	def is_unstructured(self):
		'''True if the coordinates are not structured, False otherwise.
		'''
		return not self.is_separated

	def reverse(self):
		'''Reverse the ordering of points in-place.
		'''
		raise NotImplementedError()

	@property
	def size(self):
		'''The number of points.
		'''
		raise NotImplementedError()

	def __len__(self):
		'''The number of dimensions.
		'''
		raise NotImplementedError()

	@classmethod
	def _add_coordinate_type(cls, coordinate_type, coordinate_class):
		cls._coordinate_types[coordinate_type] = coordinate_class

class UnstructuredCoords(Coords):
	'''An unstructured list of points.

	Parameters
	----------
	coords : list or tuple
		A tuple of a list of positions for each dimension.
	'''
	def __init__(self, coords):
		self.coords = [np.array(c) for c in coords]

	@classmethod
	def from_dict(cls, tree):
		'''Make an UnstructuredCoords from a dictionary, previously created by `to_dict()`.

		Parameters
		----------
		tree : dictionary
			The dictionary from which to make a new UnstructuredCoords object.

		Returns
		-------
		UnstructuredCoords
			The created object.

		Raises
		------
		ValueError
			If the dictionary is not formatted correctly.
		'''
		if tree['type'] != 'unstructured':
			raise ValueError('The type of coordinates should be "unstructured".')

		return cls(tree['coords'])

	def to_dict(self):
		'''Convert the object to a dictionary for serialization.

		Returns
		-------
		dictionary
			The created dictionary.
		'''
		tree = {
			'type': 'unstructured',
			'coords': self.coords
		}

		return tree

	@property
	def size(self):
		'''The number of points.
		'''
		return self.coords[0].size

	def __len__(self):
		'''The number of dimensions.
		'''
		return len(self.coords)

	def __getitem__(self, i):
		'''The `i`-th point for these coordinates.
		'''
		return self.coords[i]

	def __iadd__(self, b):
		'''Add `b` to the coordinates separately in-place.
		'''
		b = np.ones(len(self.coords)) * b
		for i in range(len(self.coords)):
			self.coords[i] += b[i]
		return self

	def __imul__(self, f):
		'''Multiply each coordinate with `f` separately in-place.
		'''
		f = np.ones(len(self.coords)) * f
		for i in range(len(self.coords)):
			self.coords[i] *= f[i]
		return self

	def reverse(self):
		'''Reverse the ordering of points in-place.
		'''
		for i in range(len(self.coords)):
			self.coords[i] = self.coords[i][::-1]
		return self

class SeparatedCoords(Coords):
	'''A list of points that are separable along each dimension.

	The actual points are given by the iterated tensor product of the `separated_coords`.

	Parameters
	----------
	separated_coords : list or tuple
		A tuple of a list of coordinates along each dimension.

	Attributes
	----------
	separated_coords
		A tuple of a list of coordinates along each dimension.
	'''
	def __init__(self, separated_coords):
		# Make a copy to avoid modification from outside the class
		self.separated_coords = [copy.deepcopy(s) for s in separated_coords]

	@classmethod
	def from_dict(cls, tree):
		'''Make an SeparatedCoords from a dictionary, previously created by `to_dict()`.

		Parameters
		----------
		tree : dictionary
			The dictionary from which to make a new SeparatedCoords object.

		Returns
		-------
		SeparatedCoords
			The created object.

		Raises
		------
		ValueError
			If the dictionary is not formatted correctly.
		'''
		if tree['type'] != 'separated':
			raise ValueError('The type of coordinates should be "separated".')

		return cls(tree['separated_coords'])

	def to_dict(self):
		'''Convert the object to a dictionary for serialization.

		Returns
		-------
		dictionary
			The created dictionary.
		'''
		tree = {
			'type': 'separated',
			'separated_coords': self.separated_coords
		}

		return tree

	def __getitem__(self, i):
		'''The `i`-th point for these coordinates.
		'''
		s0 = (1,) * len(self)
		j = len(self) - i - 1
		output = self.separated_coords[i].reshape(s0[:j] + (-1,) + s0[j + 1:])
		return np.broadcast_to(output, self.shape).ravel()

	@property
	def size(self):
		'''The number of points.
		'''
		return np.prod(self.shape)

	def __len__(self):
		'''The number of dimensions.
		'''
		return len(self.separated_coords)

	@property
	def dims(self):
		'''The number of points along each dimension.
		'''
		return np.array([len(c) for c in self.separated_coords])

	@property
	def shape(self):
		'''The shape of an ``numpy.ndarray`` with the right dimensions.
		'''
		return self.dims[::-1]

	def __iadd__(self, b):
		'''Add `b` to the coordinates separately in-place.
		'''
		for i in range(len(self)):
			self.separated_coords[i] += b[i]
		return self

	def __imul__(self, f):
		'''Multiply each coordinate with `f` separately in-place.
		'''
		if np.isscalar(f):
			for i in range(len(self)):
				self.separated_coords[i] *= f
		else:
			for i in range(len(self)):
				self.separated_coords[i] *= f[i]
		return self

	def reverse(self):
		'''Reverse the ordering of points in-place.
		'''
		for i in range(len(self)):
			self.separated_coords[i] = self.separated_coords[i][::-1]
		return self

class RegularCoords(Coords):
	'''A list of points that have a regular spacing in all dimensions.

	Parameters
	----------
	delta : array_like
		The spacing between the points.
	dims : array_like
		The number of points along each dimension.
	zero : array_like
		The coordinates for the first point.

	Attributes
	----------
	delta
		The spacing between the points.
	dims
		The number of points along each dimension.
	zero
		The coordinates for the first point.
	'''
	def __init__(self, delta, dims, zero=None):
		if np.isscalar(dims):
			self.dims = np.array([dims]).astype('int')
		else:
			self.dims = np.array(dims).astype('int')

		if np.isscalar(delta):
			self.delta = np.array([delta]*len(self.dims))
		else:
			self.delta = np.array(delta)

		if zero is None:
			self.zero = np.zeros(len(self.dims))
		elif np.isscalar(zero):
			self.zero = np.array([zero]*len(self.dims))
		else:
			self.zero = np.array(zero)

	@classmethod
	def from_dict(cls, tree):
		'''Make an RegularCoords from a dictionary, previously created by `to_dict()`.

		Parameters
		----------
		tree : dictionary
			The dictionary from which to make a new RegularCoords object.

		Returns
		-------
		RegularCoords
			The created object.

		Raises
		------
		ValueError
			If the dictionary is not formatted correctly.
		'''
		if tree['type'] != 'regular':
			raise ValueError('The type of coordinates should be "regular".')

		return cls(tree['delta'], tree['dims'], tree['zero'])

	def to_dict(self):
		'''Convert the object to a dictionary for serialization.

		Returns
		-------
		dictionary
			The created dictionary.
		'''
		tree = {
			'type': 'regular',
			'delta': self.delta.tolist(),
			'dims': self.dims.tolist(),
			'zero': self.zero.tolist()
		}

		return tree

	@property
	def separated_coords(self):
		'''A tuple of a list of the values for each dimension.

		The actual points are the iterated tensor product of this tuple.
		'''
		return [np.arange(n) * delta + zero for delta, n, zero in zip(self.delta, self.dims, self.zero)]

	@property
	def regular_coords(self):
		'''The tuple `(delta, dims, zero)` of the regularly-spaced coordinates.
		'''
		return self.delta, self.dims, self.zero

	@property
	def size(self):
		'''The number of points.
		'''
		return np.prod(self.dims)

	def __len__(self):
		'''The number of dimensions.
		'''
		return len(self.dims)

	@property
	def shape(self):
		'''The shape of an ``numpy.ndarray`` with the right dimensions.
		'''
		return self.dims[::-1]

	def __getitem__(self, i):
		'''The `i`-th point for these coordinates.
		'''
		s0 = (1,) * len(self)
		j = len(self) - i - 1
		t = s0[:j] + (-1,) + s0[j + 1:]
		output = self.separated_coords[i].reshape(t)
		return np.broadcast_to(output, self.shape).ravel()

	def __iadd__(self, b):
		'''Add `b` to the coordinates separately in-place.
		'''
		self.zero += b
		return self

	def __imul__(self, f):
		'''Multiply each coordinate with `f` separately in-place.
		'''
		self.delta *= f
		self.zero *= f
		return self

	def reverse(self):
		'''Reverse the ordering of points in-place.
		'''
		maximum = self.zero + self.delta * (self.dims - 1)
		self.delta = -self.delta
		self.zero = maximum
		return self

Coords._add_coordinate_type('unstructured', UnstructuredCoords)
Coords._add_coordinate_type('separated', SeparatedCoords)
Coords._add_coordinate_type('regular', RegularCoords)



import numpy as np
from ..field import Field

class FourierTransform(object):
	'''The base class for all Fourier transform implementations.

	Attributes
	----------
	input_grid : Grid
		The grid that is expected for the input field.
	output_grid : Grid
		The grid that is produced by the Fourier transform.
	'''
	def forward(self, field):
		'''Returns the forward Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to Fourier transform.

		Returns
		--------
		Field
			The Fourier transform of the field.
		'''
		raise NotImplementedError()

	def backward(self, field):
		'''Returns the inverse Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to inverse Fourier transform.

		Returns
		--------
		Field
			The inverse Fourier transform of the field.
		'''
		raise NotImplementedError()

	def get_transformation_matrix_forward(self):
		'''Returns the transformation matrix corresonding to the
		Fourier transform.

		Returns
		--------
		ndarray
			A matrix representing the Fourier transform.
		'''
		coords_in = self.input_grid.as_('cartesian').coords
		coords_out = self.output_grid.as_('cartesian').coords

		A = np.exp(-1j * np.dot(np.array(coords_out).T, coords_in))
		A *= self.input_grid.weights

		return A

	def get_transformation_matrix_backward(self):
		'''Returns the transformation matrix corresonding to the
		Fourier transform.

		Returns
		--------
		ndarray
			A matrix representing the Fourier transform.
		'''
		coords_in = self.input_grid.as_('cartesian').coords
		coords_out = self.output_grid.as_('cartesian').coords

		A = np.exp(1j * np.dot(np.array(coords_in).T, coords_out))
		A *= self.output_grid.weights
		A /= (2*np.pi)**self.input_grid.ndim

		return A

def _time_it(function, t_max=1, n_max=100):
	import time

	start = time.time()
	times = []

	while (time.time() < start + t_max) and (len(times) < n_max):
		t1 = time.time()
		function()
		t2 = time.time()
		times.append(t2 - t1)

	return np.median(times)

def make_fourier_transform(input_grid, output_grid=None, q=1, fov=1, planner='estimate'):
	'''Construct a FourierTransform object.

	The most time-efficient Fourier transform method will be chosen according to actual or estimated performance.

	Parameters
	----------
	input_grid : Grid
		The grid that will be used for the Field passed to the Fourier transform.
	output_grid : None or Grid
		The grid of the resulting field. If it is None, a optimal grid will be chosen, according to `q` and `fov`.
	q : scalar
		Describes how many samples to take in the Fourier domain. A value of 1 means critcally sampled in the Fourier domain.
	fov : scalar
		Describes how far out the Fourier domain extends. A value of 1 means the same amount of samples as the spatial domain.
	planner : string
		If it is 'estimate', performance of the different methods will be estimated from theoretical complexity estimates.
		If it is 'measure', actual Fourier transforms will be performed to get the actual performance. The latter takes longer,
		but is more accurate.

	Returns
	-------
	FourierTransform
		The Fourier transform that was requested.
	'''
	from .fast_fourier_transform import FastFourierTransform, make_fft_grid
	from .matrix_fourier_transform import MatrixFourierTransform
	from .naive_fourier_transform import NaiveFourierTransform

	if output_grid is None:
		# Choose between FFT and MFT
		if not (input_grid.is_regular and input_grid.is_('cartesian')):
			raise ValueError('For non-regular non-cartesian Grids, a Fourier transform is required to have an output_grid.')

		if input_grid.ndim not in [1,2]:
			method = 'fft'
		else:
			output_grid = make_fft_grid(input_grid, q, fov)

			if planner == 'estimate':
				# Estimate analytically from complexities
				N_in = input_grid.shape * q
				N_out = output_grid.shape

				if input_grid.ndim == 1:
					fft = 4 * N_in[0] * np.log2(N_in)
					mft = 4 * input_grid.size * N_out[0]
				else:
					fft = 4 * np.prod(N_in) * np.log2(np.prod(N_in))
					mft = 4 * (np.prod(input_grid.shape) * N_out[1] + np.prod(N_out) * input_grid.shape[0])
				if fft > mft:
					method = 'mft'
				else:
					method = 'fft'
			elif planner == 'measure':
				# Measure directly
				fft = FastFourierTransform(input_grid, q, fov)
				mft = MatrixFourierTransform(input_grid, output_grid)

				a = input_grid.zeros(dtype='complex')
				fft_time = _time_it(lambda: fft.forward(a))
				mft_time = _time_it(lambda: mft.forward(a))

				if fft_time > mft_time:
					method = 'mft'
				else:
					method = 'fft'
	else:
		# Choose between MFT and Naive
		if input_grid.is_separated and input_grid.is_('cartesian') and output_grid.is_separated and output_grid.is_('cartesian') and input_grid.ndim in [1,2]:
			method = 'mft'
		else:
			method = 'naive'

	# Make the Fourier transform
	if method == 'fft':
		return FastFourierTransform(input_grid, q, fov)
	elif method == 'mft':
		return MatrixFourierTransform(input_grid, output_grid)
	elif method == 'naive':
		return NaiveFourierTransform(input_grid, output_grid)

def multiplex_for_tensor_fields(func):
	'''A decorator for automatically multiplexing a function over the tensor directions.

	This function is used internally for simplifying the implementation of the Fourier transforms.

	Parameters
	----------
	func : function
		The function to multiplex. This function gets called for each of the tensor elements.
	'''
	def inner(self, field):
		if field.is_scalar_field:
			return func(self, field)
		else:
			f = field.reshape((-1,field.grid.size))
			res = [func(self, ff) for ff in f]
			new_shape = np.concatenate((field.tensor_shape, [-1]))
			return Field(np.array(res).reshape(new_shape), res[0].grid)

	return inner

def _get_float_and_complex_dtype(dtype):
	'''Return the floating point and complex numpy data types with the same
	bit depth per component.

	Parameters
	----------
	dtype : numpy dtype
		The dtype to convert to a floating point and complex dtype.

	Returns
	-------
	float_dtype : numpy dtype
		The floating point dtype.
	complex_dtype : numpy dtype
		The complex dtype.
	'''
	if dtype == np.dtype('float32') or dtype == np.dtype('complex64'):
		complex_dtype = 'complex64'
		float_dtype = 'float32'
	else:
		complex_dtype = 'complex128'
		float_dtype = 'float64'

	return float_dtype, complex_dtype



__all__ = [
    'circular_aperture',
    'elliptical_aperture',
    'rectangular_aperture',
    'regular_polygon_aperture',
    'hexagonal_aperture',
    'make_spider',
    'make_spider_infinite',
    'make_obstructed_circular_aperture',
    'make_obstruction',
    'make_rotated_aperture',
    'make_segmented_aperture',
    'make_hexagonal_segmented_aperture',
    'make_magellan_aperture',
    'make_luvoir_a_aperture',
    'make_luvoir_a_lyot_stop',
    'make_hicat_aperture',
    'make_hicat_lyot_stop',
	'make_vlt_aperture'
]

from .generic import *
from .realistic import *



from .mode_basis import ModeBasis
from ..field import Field
import numpy as np

_epsilon = 1e-12

def make_cosine_basis(grid, fourier_grid, sort_by_energy=True):
	'''Make a cosine basis.

	Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

	Parameters
	----------
	grid : Grid
		The :class:`Grid` on which to calculate the modes.
	fourier_grid : Grid
		The grid defining all frequencies.
	sort_by_energy : bool
		Whether to sort by increasing energy or not.

	Returns
	-------
	ModeBasis
		The mode basis containing all cosine modes.
	'''
	modes = []
	energies = []
	ignore_list = []

	c = np.array(grid.coords)

	for i, p in enumerate(fourier_grid.points):
		if i in ignore_list:
			continue

		mode = Field(np.cos(np.dot(p, c)), grid)

		# Correctly normalize the non-zero energy mode.
		if not np.dot(p, p) < _epsilon**2:
			mode *= np.sqrt(2)

		modes.append(mode)

		j = fourier_grid.closest_to(-p)

		dist = fourier_grid.points[j] + p
		dist2 = np.dot(dist, dist)

		p_length2 = np.dot(p, p)
		energies.append(p_length2)

		if dist2 < (_epsilon * p_length2):
			ignore_list.append(j)

	if sort_by_energy:
		ind = np.argsort(energies)
		modes = [modes[i] for i in ind]

	return ModeBasis(modes, grid)

def make_sine_basis(grid, fourier_grid, sort_by_energy=True):
	'''Make a sine basis.

	Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

	Parameters
	----------
	grid : Grid
		The :class:`Grid` on which to calculate the modes.
	fourier_grid : Grid
		The grid defining all frequencies.
	sort_by_energy : bool
		Whether to sort by increasing energy or not.

	Returns
	-------
	ModeBasis
		The mode basis containing all sine modes.
	'''
	modes = []
	energies = []
	ignore_list = []

	c = np.array(grid.coords)

	for i, p in enumerate(fourier_grid.points):
		if i in ignore_list:
			continue

		mode = Field(np.sin(np.dot(p, c)) * np.sqrt(2), grid)

		# Filter out the zero energy mode.
		if np.dot(p, p) < _epsilon**2:
			continue

		modes.append(mode)

		j = fourier_grid.closest_to(-p)

		dist = fourier_grid.points[j] + p
		dist2 = np.dot(dist, dist)

		p_length2 = np.dot(p, p)
		energies.append(p_length2)

		if dist2 < (_epsilon * p_length2):
			ignore_list.append(j)

	if sort_by_energy:
		ind = np.argsort(energies)
		modes = [modes[i] for i in ind]

	return ModeBasis(modes, grid)

def make_fourier_basis(grid, fourier_grid, sort_by_energy=True):
	'''Make a Fourier basis.

	Fourier modes this function are defined to be real. This means that for each point, both a sine and cosine mode is returned.

	Repeated frequencies will not be repeated in this mode basis. This means that opposite points in the `fourier_grid` will be silently ignored.

	Parameters
	----------
	grid : Grid
		The :class:`Grid` on which to calculate the modes.
	fourier_grid : Grid
		The grid defining all frequencies.
	sort_by_energy : bool
		Whether to sort by increasing energy or not.

	Returns
	-------
	ModeBasis
		The mode basis containing all Fourier modes.
	'''
	modes_cos = []
	modes_sin = []
	energies = []
	ignore_list = []

	c = np.array(grid.coords)

	for i, p in enumerate(fourier_grid.points):
		if i in ignore_list:
			continue

		mode_cos = Field(np.cos(np.dot(p, c)), grid)
		mode_sin = Field(np.sin(np.dot(p, c)), grid)

		modes_cos.append(mode_cos)
		modes_sin.append(mode_sin)

		j = fourier_grid.closest_to(-p)

		dist = fourier_grid.points[j] + p
		dist2 = np.dot(dist, dist)

		p_length2 = np.dot(p, p)
		energies.append(p_length2)

		if dist2 < (_epsilon * p_length2):
			ignore_list.append(j)

	if sort_by_energy:
		ind = np.argsort(energies)
		modes_sin = [modes_sin[i] for i in ind]
		modes_cos = [modes_cos[i] for i in ind]
		energies = np.array(energies)[ind]

	modes = []
	for i, E in enumerate(energies):
		# Filter out and correctly normalize zero energy vs non-zero energy modes.
		if E > _epsilon:
			modes.append(modes_cos[i] * np.sqrt(2))
			modes.append(modes_sin[i] * np.sqrt(2))
		else:
			modes.append(modes_cos[i])

	return ModeBasis(modes, grid)

def make_complex_fourier_basis(grid, fourier_grid, sort_by_energy=True):
	'''Make a complex Fourier basis.

	Fourier modes this function are defined to be complex. For each point in `fourier_grid` the complex Fourier mode is contained in the output.

	Parameters
	----------
	grid : Grid
		The :class:`Grid` on which to calculate the modes.
	fourier_grid : Grid
		The grid defining all frequencies.
	sort_by_energy : bool
		Whether to sort by increasing energy or not.

	Returns
	-------
	ModeBasis
		The mode basis containing all Fourier modes.
	'''
	c = np.array(grid.coords)

	modes = [Field(np.exp(1j * np.dot(p, c)), grid) for p in fourier_grid.points]
	energies = [np.dot(p, p) for p in fourier_grid.points]

	if sort_by_energy:
		ind = np.argsort(energies)
		modes = [modes[i] for i in ind]

	return ModeBasis(modes, grid)



from hcipy import *
import numpy as np
import scipy
import matplotlib.pyplot as plt

def test_zernike_modes():
	grid = make_pupil_grid(256)
	aperture_mask = circular_aperture(1)(grid) > 0

	modes = make_zernike_basis(200, 1, grid)

	assert np.abs(np.std(modes[0][aperture_mask])) < 2e-2

	for m in modes[1:]:
		assert np.abs(np.std(m[aperture_mask]) - 1) < 2e-2

	for i, m in enumerate(modes):
		zn, zm = noll_to_zernike(i+1)
		assert np.allclose(m, zernike(zn, zm, grid=grid))

def test_zernike_indices():
	for i in range(1,500):
		n, m = noll_to_zernike(i)
		assert i == zernike_to_noll(n, m)

		n, m = ansi_to_zernike(i)
		assert i == zernike_to_ansi(n, m)

def test_disk_harmonic_modes():
	grid = make_pupil_grid(128)
	aperture_mask = circular_aperture(1)(grid) > 0

	num_modes = 20

	for bc in ['dirichlet', 'neumann']:
		modes = make_disk_harmonic_basis(grid, num_modes, bc=bc)

		for i, m1 in enumerate(modes):
			for j, m2 in enumerate(modes):
				product = np.sum((m1 * m2)[aperture_mask])
				assert np.abs(product - np.eye(num_modes)[i,j]) < 1e-2

def test_LP_modes():
	grid = make_pupil_grid(128)

	# Test for single-mode
	modes = make_LP_modes(grid, 2.4, 0.1, return_betas=False)
	assert len(modes) == 1

	# Test orthogonality
	modes = make_LP_modes(grid, 25, 0.1, return_betas=False)
	for i, m1 in enumerate(modes):
		for j, m2 in enumerate(modes):
			product = np.real(np.sum(m1 * m2.conj() * grid.weights))
			assert np.abs(product - 1 if i == j else 0) <= 1e-2

def test_sparse_mode_basis():
	transformation_matrix = np.empty((100 * 100, 100))
	for i in range(100):
		transformation_matrix[:, i] = np.random.randn(100 * 100)
		transformation_matrix[np.random.choice(100 * 100, 80 * 100, False), i] = 0

	mode_basis_1 = ModeBasis(transformation_matrix)

	assert mode_basis_1.is_dense
	assert not mode_basis_1.is_sparse
	assert not scipy.sparse.issparse(mode_basis_1.transformation_matrix)

	mode_basis_2 = mode_basis_1.to_dense(copy=False)

	assert mode_basis_2 is mode_basis_1

	mode_basis_3 = mode_basis_1.to_dense(copy=True)

	assert mode_basis_3 is not mode_basis_1
	assert np.allclose(mode_basis_1.transformation_matrix, mode_basis_3.transformation_matrix)

	mode_basis_4 = mode_basis_1.to_sparse()

	assert mode_basis_4.is_sparse
	assert not mode_basis_4.is_dense
	assert scipy.sparse.issparse(mode_basis_4.transformation_matrix)
	assert mode_basis_4.transformation_matrix.nnz < np.prod(mode_basis_4.transformation_matrix.shape)

	mode_basis_5 = mode_basis_4.to_dense()

	assert mode_basis_5.is_dense
	assert not mode_basis_5.is_sparse
	assert np.allclose(mode_basis_1.transformation_matrix, mode_basis_5.transformation_matrix)
	assert not scipy.sparse.issparse(mode_basis_5.transformation_matrix)

def test_gaussian_laguerre_modes():
	grid = make_focal_grid(32, 4)

	p_max = 5
	l_max = 5
	mode_field_diameter = 1

	modes = make_gaussian_laguerre_basis(grid, p_max, l_max, mode_field_diameter)
	num_modes = len(modes)

	for i, m1 in enumerate(modes):
		for j, m2 in enumerate(modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-6

def test_gaussian_hermite_modes():
	grid = make_focal_grid(32, 4)

	num_modes = 50
	mode_field_diameter = 1

	modes = make_gaussian_hermite_basis(grid, num_modes, mode_field_diameter)

	for i, m1 in enumerate(modes):
		for j, m2 in enumerate(modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-6

def test_fourier_modes():
	grid = make_pupil_grid(32)
	aperture_mask = circular_aperture(1)(grid) > 0

	fourier_grid = make_fft_grid(grid, 1, 0.2)

	# Cosine modes
	cosine_modes = make_cosine_basis(grid, fourier_grid)
	num_modes = len(cosine_modes)

	for i, m1 in enumerate(cosine_modes):
		for j, m2 in enumerate(cosine_modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

	# Sine modes
	sine_modes = make_sine_basis(grid, fourier_grid)
	num_modes = len(sine_modes)

	for i, m1 in enumerate(sine_modes):
		for j, m2 in enumerate(sine_modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

	# Fourier basis
	fourier_modes = make_fourier_basis(grid, fourier_grid)
	num_modes = len(fourier_modes)

	for i, m1 in enumerate(fourier_modes):
		for j, m2 in enumerate(fourier_modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12

	# Complex Fourier basis
	complex_fourier_modes = make_complex_fourier_basis(grid, fourier_grid)
	num_modes = len(complex_fourier_modes)

	for i, m1 in enumerate(complex_fourier_modes):
		for j, m2 in enumerate(complex_fourier_modes):
			product = np.sum((m1.conj() * m2) * grid.weights).real
			assert np.abs(product - np.eye(num_modes)[i, j]) < 1e-12



from __future__ import division
import functools

import numpy as np
from ..field import Field, CartesianGrid, UnstructuredCoords, make_hexagonal_grid

def circular_aperture(diameter, center=None):
	'''Makes a Field generator for a circular aperture.

	Parameters
	----------
	diameter : scalar
		The diameter of the aperture.
	center : array_like
		The center of the aperture

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''
	if center is None:
		shift = np.zeros(2)
	else:
		shift = center * np.ones(2)

	def func(grid):
		if grid.is_('cartesian'):
			if grid.is_separated:
				x, y = grid.separated_coords
				f = ((((x - shift[0])**2)[np.newaxis, :] + ((y - shift[1])**2)[:, np.newaxis]) <= (diameter / 2)**2).ravel()
			else:
				x, y = grid.coords
				f = ((x - shift[0])**2 + (y - shift[1])**2) <= (diameter / 2)**2
		else:
			f = grid.r <= (diameter / 2)

		return Field(f.astype('float'), grid)

	return func

def elliptical_aperture(diameters, center=None):
	'''Makes a Field generator for an elliptical aperture.

	Parameters
	----------
	diameters : array_like
		The diameters of the aperture in the two directions.
	center : array_like or None
		The center of the aperture.

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''
	if center is None:
		shift = np.zeros(2)
	else:
		shift = center * np.ones(2)

	def func(grid):
		if grid.is_separated:
			x, y = grid.separated_coords
			f = ((((x - shift[0]) / (diameters[0] / 2))**2)[np.newaxis, :] + (((y - shift[1]) / (diameters[1] / 2))**2)[:, np.newaxis] <= 1).ravel()
		else:
			x, y = grid.as_('cartesian').coords
			f = (((x - shift[0]) / (diameters[0] / 2))**2 + ((y - shift[1]) / (diameters[1] / 2))**2) <= 1

		return Field(f.astype('float'), grid)

	return func

def rectangular_aperture(size, center=None):
	'''Makes a Field generator for a rectangular aperture.

	Parameters
	----------
	size : scalar or array_like
		The length of the sides. If this is scalar, a square aperture is assumed.
	center : array_like
		The center of the aperture

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''
	dim = size * np.ones(2)

	if center is None:
		shift = np.zeros(2)
	else:
		shift = center * np.ones(2)

	def func(grid):
		if grid.is_('cartesian') and grid.is_separated:
			x, y = grid.separated_coords
			f = ((np.abs(x - shift[0]) <= (dim[0] / 2))[np.newaxis, :] * (np.abs(y - shift[1]) <= (dim[1] / 2))[:, np.newaxis]).ravel()
		else:
			x, y = grid.as_('cartesian').coords
			f = (np.abs(x - shift[0]) <= (dim[0] / 2)) * (np.abs(y - shift[1]) <= (dim[1] / 2))

		return Field(f.astype('float'), grid)

	return func

def regular_polygon_aperture(num_sides, circum_diameter, angle=0, center=None):
	'''Makes a Field generator for a regular-polygon-shaped aperture.

	Parameters
	----------
	num_sides : integer
		The number of sides for the polygon.
	circum_diameter : scalar
		The circumdiameter of the polygon.
	angle : scalar
		The angle by which to turn the polygon.
	center : array_like
		The center of the aperture

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''
	if num_sides < 3:
		raise ValueError('The number of sides for a regular polygon has to greater or equal to 3.')

	if center is None:
		shift = np.zeros(2)
	else:
		shift = center * np.ones(2)

	epsilon = 1e-6

	apothem = np.cos(np.pi / num_sides) * circum_diameter / 2
	apothem += apothem * epsilon

	if center is None:
		shift = np.zeros(2)
	else:
		shift = center * np.ones(2)

	# Make use of symmetry
	if num_sides % 2 == 0:
		thetas = np.arange(int(num_sides / 2), dtype='float') * np.pi / int(num_sides / 2) + angle
	else:
		thetas = np.arange(int(num_sides / 2) + 1) * (num_sides - 2) * np.pi / (num_sides / 2) + angle

	mask = rectangular_aperture(circum_diameter)

	def func(grid):
		g = grid.as_('cartesian')

		if g.is_separated:
			x, y = g.separated_coords

			x = x - shift[0]
			y = y - shift[1]

			ind_x = np.flatnonzero(x**2 < ((circum_diameter / 2)**2))
			if not len(ind_x):
				return grid.zeros()

			ind_y = np.flatnonzero(y**2 < ((circum_diameter / 2)**2))
			if not len(ind_y):
				return grid.zeros()

			m_x = slice(ind_x[0], ind_x[-1] + 1)
			m_y = slice(ind_y[0], ind_y[-1] + 1)

			x = x[m_x]
			y = y[m_y]

			f_sub = np.ones((len(ind_y), len(ind_x)))

			if num_sides % 2 == 0:
				for theta in thetas:
					f_sub *= ((np.cos(theta) * x)[np.newaxis, :] + (np.sin(theta) * y)[:, np.newaxis])**2 <= apothem**2
			else:
				for theta in thetas:
					f_sub *= (np.abs(np.sin(theta) * x)[np.newaxis, :] - (np.cos(theta) * y)[:, np.newaxis]) <= apothem

			f = np.zeros(g.shape)
			f[m_y, m_x] = f_sub

			return Field(f.ravel(), grid)

		# Slow backup method
		f = np.ones(grid.size, dtype='float')
		m = mask(g) != 0

		x, y = g.coords
		x = x[m] - shift[0]
		y = y[m] - shift[1]

		f[~m] = 0

		# Make use of symmetry
		if num_sides % 2 == 0:
			for theta in thetas:
				f[m] *= (np.cos(theta) * x + np.sin(theta) * y)**2 <= apothem**2
		else:
			for theta in thetas:
				f[m] *= (np.abs(np.sin(theta) * x) + -np.cos(theta) * y) <= apothem

		return Field(f, grid)

	return func

# Convenience function
def hexagonal_aperture(circum_diameter, angle=0, center=None):
	'''Makes a Field generator for a hexagon aperture.

	Parameters
	----------
	circum_diameter : scalar
		The circumdiameter of the polygon.
	angle : scalar
		The angle by which to turn the hexagon.
	center : array_like
		The center of the aperture

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''
	return regular_polygon_aperture(6, circum_diameter, angle, center)

def make_spider(p1, p2, spider_width):
	'''Make a rectangular obstruction from `p1` to `p2`.

	Parameters
	----------
	p1 : list or ndarray
		The starting coordinates of the spider.
	p2 : list or ndarray
		The end coordinates of the spider.
	spider_width : scalar
		The full width of the spider.

	Returns
	-------
	Field generator
		The spider obstruction.
	'''
	delta = np.array(p2) - np.array(p1)
	shift = delta / 2 + np.array(p1)

	spider_angle = np.arctan2(delta[1], delta[0])
	spider_length = np.linalg.norm(delta)

	spider = rectangular_aperture((spider_length, spider_width))

	def func(grid):
		x, y = grid.shifted(-shift).coords

		x_new = x * np.cos(spider_angle) + y * np.sin(spider_angle)
		y_new = y * np.cos(spider_angle) - x * np.sin(spider_angle)
		spider_grid = CartesianGrid(UnstructuredCoords([x_new, y_new]))

		return 1 - spider(spider_grid)
	return func

def make_spider_infinite(p, angle, spider_width):
	'''Make an infinite spider starting at `p` and extending at an angle `angle`.

	Parameters
	----------
	p : list or ndarray
		The starting coordinate of the spider.
	angle : scalar
		The angle to which the spider is pointing in degrees.
	spider_width : scalar
		The full width of the spider.

	Returns
	-------
	Field generator
		The spider obstruction.
	'''
	spider_angle = np.radians(angle)

	def func(grid):
		x,y = grid.shifted(p).coords

		x_new = x * np.cos(spider_angle) + y * np.sin(spider_angle)
		y_new = y * np.cos(spider_angle) - x * np.sin(spider_angle)
		infinite_spider = np.logical_and(np.abs(y_new) <= (spider_width / 2), x_new >= 0)

		return Field(1 - infinite_spider, grid)
	return func

def make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio, num_spiders=0, spider_width=0.01):
	'''Make a simple circular aperture with central obscuration and support structure.

	Parameters
	----------
	pupil_diameter : scalar
		The diameter of the circular aperture.
	central_obscuration_ratio : scalar
		The ratio of the diameter of the central obscuration compared to the pupil diameter.
	num_spiders : int
		The number of spiders holding up the central obscuration.
	spider_width : scalar
		The full width of the spiders.

	Returns
	-------
	Field generator
		The circularly obstructed aperture.
	'''
	central_obscuration_diameter = pupil_diameter * central_obscuration_ratio

	def func(grid):
		pupil_outer = circular_aperture(pupil_diameter)(grid)
		pupil_inner = circular_aperture(central_obscuration_diameter)(grid)
		spiders = 1

		spider_angles = np.linspace(0, 2*np.pi, num_spiders, endpoint=False)

		for angle in spider_angles:
			x = pupil_diameter * np.cos(angle)
			y = pupil_diameter * np.sin(angle)

			spiders *= make_spider((0, 0), (x, y), spider_width)(grid)

		return (pupil_outer - pupil_inner) * spiders
	return func

def make_obstruction(aperture):
	'''Create an obstruction of `aperture`.

	Parameters
	----------
	aperture : Field generator
		The aperture to invert.

	Returns
	-------
	Field generator
		The obstruction.
	'''
	return lambda grid: 1 - aperture(grid)

def make_rotated_aperture(aperture, angle):
	'''Create a rotated version of `aperture`.

	Parameters
	----------
	aperture : Field generator
		The aperture to rotate.
	angle : scalar
		The angle in radians by which to rotate the aperture.

	Returns
	-------
	Field generator
		The rotated aperture.
	'''
	return lambda grid: Field(aperture(grid.rotated(-angle)), grid)

def make_segmented_aperture(segment_shape, segment_positions, segment_transmissions=1, return_segments=False):
	'''Create a segmented aperture.

	Parameters
	----------
	segment_shape : Field generator
		The shape for each of the segments.
	segment_positions : Grid
		The center position for each of the segments.
	segment_transmissions : scalar or ndarray
		The transmission for each of the segments. If this is a scalar, the same transmission is used for all segments.
	return_segments : boolean
		Whether to return a ModeBasis of all segments as well.

	Returns
	-------
	Field generator
		The segmented aperture.
	list of Field generators
		The segments. Only returned if return_segments is True.
	'''
	segment_transmissions = np.ones(segment_positions.size) * segment_transmissions

	def func(grid):
		res = np.zeros(grid.size, dtype=segment_transmissions.dtype)

		for p, t in zip(segment_positions.points, segment_transmissions):
			segment = segment_shape(grid.shifted(-p))
			res[segment > 0.5] = t

		return Field(res, grid)

	if return_segments:
		def seg(grid, p, t):
			return segment_shape(grid.shifted(-p)) * t

		segments = []
		for p, t in zip(segment_positions.points, segment_transmissions):
			segments.append(functools.partial(seg, p=p, t=t))

		return func, segments
	else:
		return func

def make_hexagonal_segmented_aperture(num_rings, segment_flat_to_flat, gap_size, starting_ring=0, return_segments=False):
	'''Create a hexagonal segmented aperture.

	All segments have a flat-to-flat size of `segment_flat_to_flat`, which means that
	the segment pitch (the distance between the centers of adjacent segments) is the sum
	of the segment flat-to-flat size and the gap size.

	Parameters
	----------
	num_rings : int
		The number of rings of hexagons to include, not counting the central segment.
		So 2 for a JWST-like aperture, 3 for a Keck-like aperture, and so on.
	segment_flat_to_flat : scalar
		The distance between sides (flat-to-flat) of a single segment.
	gap_size : scalar
		The gap between adjacent segments.
	starting_ring : int
		The first ring of segments. This can be used to exclude the center segment (by
		setting it to one), or the center segment and first ring (by setting it to two).
		The default (zero) includes the center segment.
	return_segments : boolean
		Whether to return a ModeBasis of all segments as well.

	Returns
	-------
	Field generator
		The segmented aperture.
	list of Field generators
		The segments. Only returned if return_segments is True.
	'''
	segment_circum_diameter = segment_flat_to_flat * 2 / np.sqrt(3)
	segment = hexagonal_aperture(segment_circum_diameter, np.pi / 2)

	segment_pitch = segment_flat_to_flat + gap_size
	segment_positions = make_hexagonal_grid(segment_pitch, num_rings, pointy_top=False)

	if starting_ring != 0:
		starting_segment = 3 * (starting_ring - 1) * starting_ring + 1

		mask = segment_positions.zeros(dtype='bool')
		mask[starting_segment:] = True

		segment_positions = segment_positions.subset(mask)

	return make_segmented_aperture(segment, segment_positions, return_segments=return_segments)



from .infinite_atmospheric_layer import InfiniteAtmosphericLayer

import numpy as np

def make_standard_atmospheric_layers(input_grid, L0=10):
	heights = np.array([500, 1000, 2000, 4000, 8000, 16000])
	velocities = np.array([10, 10, 10, 10, 10, 10])
	Cn_squared = np.array([0.2283, 0.0883, 0.0666, 0.1458, 0.3350, 0.1350]) * 1e-12

	layers = []
	for h, v, cn in zip(heights, velocities, Cn_squared):
		layers.append(InfiniteAtmosphericLayer(input_grid, cn, L0, v, h, 2))

	return layers



import numpy as np
from math import sqrt
from scipy.special import eval_genlaguerre

def gaussian_laguerre(p, l, mode_field_diameter=1, grid=None):
	r'''Creates a Gaussian-Laguerre mode.

	This function evaluates a (p,l) order Gaussian-Laguerre mode on a grid.
	The definition of the modes are the following,

	.. math::
		\exp{\left(-\frac{r^2}{w_0^2}\right)} L_p^{|l|}\left(\frac{2r^2}{w_0^2} \right) \left(\sqrt{2}\frac{r}{w_0}\right)

	Here :math:`w_0` is the mode_field_radius, which is :math:`\mathrm{MFD}/2`.
	And :math:`L_p^{|l|}` are the generalized Laguerre Polynomials.
	All modes are numerical normalized to have a total power of 1.

	More details on generalized Laguerre Polynomials can be found on:
	http://mathworld.wolfram.com/AssociatedLaguerrePolynomial.html

	Parameters
	----------
	p : int
		The radial order.
	l : int
		The azimuthal order.
	mode_field_diameter : scalar
		The mode field diameter of the mode.
	grid : Grid
		The grid on which to evaluate the mode.

	Returns
	-------
	Field
		The evaluated mode.
	'''
	from ..field import Field

	if grid.is_separated and grid.is_('polar'):
		R, Theta = grid.separated_coords
	else:
		R, Theta = grid.as_('polar').coords

	# Easy access
	r = 2 * R / mode_field_diameter
	r2 = r**2

	# Compute the mode
	lg = (r * sqrt(2))**abs(l) * np.exp(-r2) * np.exp(-1j * l * Theta) * eval_genlaguerre(p, abs(l), 2 * r2)

	# Numerically normalize the mode
	lg /= np.sqrt(np.sum(np.abs(lg)**2 * grid.weights))

	return Field(lg, grid)

# High level functions
def make_gaussian_laguerre_basis(grid, pmax, lmax, mode_field_diameter, pmin=0):
	'''Creates a Gaussian-Laguerre mode basis.

	This function evaluates Gaussian-Laguerre modes. For each radial order
	within [pmin, pmax] it will calculate the azimuthal order [-lmax, lmax] inclusive.
	This function returns a ModeBasis made out of these Gaussian-Laguerre modes.

	Parameters
	----------
	grid : Grid
		The grid on which to evaluate the Gaussian-Laguerre mode.
	pmax : int
		The maximum radial order of the modes.
	lmax : int
		The maximum azimuthal order.
	mode_field_diameter : scalar
		The mode field diameter of the Gaussian-Laguerre mode.
	pmin : int
		The minimal radial order.

	Returns
	-------
	ModeBasis
		The Gaussian-Laguerre modes.
	'''
	from .mode_basis import ModeBasis

	modes = [gaussian_laguerre(pi, li, mode_field_diameter, grid) for li in range(-lmax, lmax + 1) for pi in range(pmin, pmax)]
	return ModeBasis(modes, grid)



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..optics import PhaseApodizer, LinearRetarder, Wavefront
from ..aperture import circular_aperture
from ..field import make_uniform_grid, Field

import numpy as np

class ZernikeWavefrontSensorOptics(WavefrontSensorOptics):
	'''The optical elements for a Zernike wavefront sensor.

	This class uses a propagation scheme of [NDiaye2013]_, similar to that of Lyot coronagraphs
	with a small blocking mask, see [Soummer2007]_.

	.. [NDiaye2013] N'Diaye et al. 2013, "Calibration of quasi-static
		aberrations in exoplanet direct-imaging instruments with a
		Zernike phase-mask sensor"

	.. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
		coronagraph propagation"

	Parameters
	----------
	input_grid : Grid
		The grid on which the input wavefront is defined.
	phase_step : scalar
		The phase of the phase dot of the zernike wavefront sensor. The default is pi/2.
	phase_dot_diameter : scalar
		The diameter of the phase dot. This has units of lambda_0/D.
	num_pix : scalar
		The number of pixels across the phase dot.
	pupil_diameter : scalar
		The diameter of the pupil. This is used for calculating the size of the phase dot.
	reference_wavelength : scalar
		The reference wavelength. This is used for calcualting the size of the phase dot.
	'''
	def __init__(self, input_grid, phase_step=np.pi / 2, phase_dot_diameter=1.06, num_pix=128, pupil_diameter=1, reference_wavelength=1):
		self.input_grid = input_grid
		self.output_grid = input_grid

		# Make the phase dot
		phase_dot_diameter *= reference_wavelength / pupil_diameter
		focal_grid = make_uniform_grid([num_pix, num_pix], phase_dot_diameter)
		self.phase_dot = PhaseApodizer(circular_aperture(phase_dot_diameter)(focal_grid) * phase_step)

		# Make the propagator
		self.prop = FraunhoferPropagator(input_grid, focal_grid)

	def forward(self, wavefront):
		'''Propagates a wavefront through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf : Wavefront
			The output wavefront.
		'''
		wf_foc = self.prop.forward(wavefront)
		wf_foc.electric_field -= self.phase_dot.forward(wf_foc).electric_field

		pup = self.prop.backward(wf_foc)
		pup.electric_field[:] = wavefront.electric_field - pup.electric_field

		return pup

	def backward(self, wavefront):
		'''Propagates a wavefront backwards through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf : Wavefront
			The output wavefront.
		'''
		wf_foc = self.prop.forward(wavefront)
		wf_foc.electric_field -= self.phase_dot.backward(wf_foc).electric_field

		pup = self.prop.backward(wf_foc)
		pup.electric_field[:] = wavefront.electric_field - pup.electric_field

		return pup

class ZernikeWavefrontSensorEstimator(WavefrontSensorEstimator):
	'''Estimates the wavefront slopes from pyramid wavefront sensor images.

	Parameters
	----------
	aperture : function
		A function which mask the output phase measurements.
	output_grid : Grid
		The grid on which the output of a pyramid wavefront sensor is sampled.
	reference : Field
		A reference image to subtract from the Zernike wavefront sensor data.

	Attributes
	----------
	measurement_grid : Grid
		The grid on which the phase measurements are defined.
	pupil_mask : array_like
		A mask for the phase measurements.
	num_measurements : int
		The number of pixels in the output vector.
	reference : Field
		A reference image to subtract from the Zernike wavefront sensor data.
	'''
	def __init__(self, aperture, output_grid, reference):
		self.measurement_grid = output_grid
		self.pupil_mask = aperture(self.measurement_grid)
		self.reference = reference
		self.num_measurements = int(np.sum(self.pupil_mask > 0))

	def estimate(self, images):
		'''A function which estimates the phase from a Zernike wavefront sensor image.

		Parameters
		----------
		images - List
			A list of scalar intensity fields containing Zernike wavefront sensor images.

		Returns
		-------
		res - Field
			A field with phase estimates.
		'''
		image = images[0]

		intensity_measurements = (image - self.reference).ravel() * self.pupil_mask
		res = Field(intensity_measurements[self.pupil_mask > 0], self.pupil_mask.grid)
		return res

class VectorZernikeWavefrontSensorOptics(WavefrontSensorOptics):
	'''The optical elements for a vector-Zernike wavefront sensor.

	The vector Zernike is based on [Doelman2019]_.

	This class uses a propagation scheme of [NDiaye2013]_, similar to that of Lyot coronagraphs
	with a small blocking mask, see [Soummer2007]_. The reference wavefront is also propagated
	through a half-wave retarder because the vector-Zernike mask is a half-wave
	retarder.

	.. [Doelman2019] Doelman et al. 2019, "Simultaneous phase and amplitude aberration sensing
		with a liquid-crystal vector-Zernike phase mask"

	.. [NDiaye2013] N'Diaye et al. 2013, "Calibration of quasi-static
		aberrations in exoplanet direct-imaging instruments with a
		Zernike phase-mask sensor"

	.. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
		coronagraph propagation"

	Parameters
	----------
	input_grid : Grid
		The grid on which the input wavefront is defined.
	phase_retardation : scalar or Field
		The relative phase retardation induced between the fast and slow axis.
	phase_step : scalar
		The phase of the phase dot of the zernike wavefront sensor. The default is pi/2.
	phase_dot_diameter : scalar
		The diameter of the phase dot. This has units of lambda_0/D.
	num_pix : scalar
		The number of pixels across the phase dot.
	pupil_diameter : scalar
		The diameter of the pupil. This is used for calculating the size of the phase dot.
	reference_wavelength : scalar
		The reference wavelength. This is used for calcualting the size of the phase dot.
	'''
	def __init__(self, input_grid, phase_retardation=np.pi, phase_step=np.pi / 2, phase_dot_diameter=1.06, num_pix=128, pupil_diameter=1, reference_wavelength=1):
		self.input_grid = input_grid
		self.output_grid = input_grid
		self._phase_retardation = phase_retardation

		# Make the vector-Zernike wavefront sensor mask
		phase_dot_diameter *= reference_wavelength / pupil_diameter
		focal_grid = make_uniform_grid([num_pix, num_pix], phase_dot_diameter)
		phase_dot = circular_aperture(phase_dot_diameter)(focal_grid) * phase_step

		self.vZWFS_mask = LinearRetarder(phase_retardation, phase_dot / 2)

		# Make half-wave plate for reference offset
		self.HWP = LinearRetarder(phase_retardation, 0)

		# Make the propagator
		self.prop = FraunhoferPropagator(input_grid, focal_grid)

	@property
	def phase_retardation(self):
		'''The phase retardation of the linear retarder
		'''
		return self._phase_retardation

	@phase_retardation.setter
	def phase_retardation(self, phase_retardation):
		self._phase_retardation = phase_retardation

		self.vZWFS_mask.phase_retardation = phase_retardation
		self.HWP.phase_retardation = phase_retardation

	def forward(self, wavefront):
		'''Propagates a wavefront through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf : Wavefront
			The output wavefront.
		'''
		wf_foc = self.prop.forward(wavefront)
		if wf_foc.is_scalar:
			wf_foc = Wavefront(wf_foc.electric_field, wavelength=wf_foc.wavelength, input_stokes_vector=(1, 0, 0, 0))

		wf_foc.electric_field = self.HWP.forward(wf_foc).electric_field - self.vZWFS_mask.forward(wf_foc).electric_field

		pup = self.prop.backward(wf_foc)
		pup.electric_field[:] = self.HWP.forward(wavefront).electric_field - pup.electric_field

		return pup

	def backward(self, wavefront):
		'''Propagates a wavefront backwards through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf : Wavefront
			The output wavefront.
		'''
		wf_foc = self.prop.forward(wavefront)
		if wf_foc.is_scalar:
			wf_foc = Wavefront(wf_foc.electric_field, wavelength=wf_foc.wavelength, input_stokes_vector=(1, 0, 0, 0))

		wf_foc.electric_field = self.HWP.backward(wf_foc).electric_field - self.vZWFS_mask.backward(wf_foc).electric_field

		pup = self.prop.backward(wf_foc)
		pup.electric_field[:] = self.HWP.backward(wavefront).electric_field - pup.electric_field

		return pup



from hcipy import *
import numpy as np

def test_vortex_coronagraph():
	pupil_grid = make_pupil_grid(256)
	focal_grid = make_focal_grid(4, 32)
	prop = FraunhoferPropagator(pupil_grid, focal_grid)

	aperture = circular_aperture(1)
	aperture = evaluate_supersampled(aperture, pupil_grid, 8)

	lyot = circular_aperture(0.99)
	lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

	for charge in [2, 4, 6, 8]:
		vortex = VortexCoronagraph(pupil_grid, charge)

		wf = Wavefront(aperture)
		wf.total_power = 1

		img_ref = prop(wf)

		wf = vortex(wf)
		wf.electric_field *= lyot
		img = prop(wf)

		assert img.total_power < 1e-6
		assert img.intensity.max() / img_ref.intensity.max() < 1e-8

def test_vector_vortex_coronagraph():
	pupil_grid = make_pupil_grid(256)
	focal_grid = make_focal_grid(4, 32)
	prop = FraunhoferPropagator(pupil_grid, focal_grid)

	aperture = circular_aperture(1)
	aperture = evaluate_supersampled(aperture, pupil_grid, 8)

	lyot = circular_aperture(0.99)
	lyot = evaluate_supersampled(lyot, pupil_grid, 8) > 1 - 1e-5

	for charge in [2, 4, 6, 8]:
		vortex = VectorVortexCoronagraph(charge)

		wf = Wavefront(aperture)
		wf.total_power = 1

		img_ref = prop(wf)

		wf = vortex(wf)
		wf.electric_field *= lyot
		img = prop(wf)

		assert img.total_power < 1e-6
		assert img.intensity.max() / img_ref.intensity.max() < 1e-8

def test_ravc():
	pupil_grid = make_pupil_grid(256)
	focal_grid = make_focal_grid(4, 32)
	prop = FraunhoferPropagator(pupil_grid, focal_grid)

	for co in [0.1, 0.2, 0.3]:
		aperture = lambda grid: circular_aperture(1)(grid) - circular_aperture(co)(grid)
		aperture = evaluate_supersampled(aperture, pupil_grid, 16)

		aper = aperture > 1e-5

		for charge in [2, 4]:
			pupil, lyot = make_ravc_masks(co, charge, lyot_undersize=0.02)

			pupil = evaluate_supersampled(pupil, pupil_grid, 4)
			lyot = evaluate_supersampled(lyot, pupil_grid, 4)

			vortex = VortexCoronagraph(pupil_grid, charge)

			wf = Wavefront(aper)
			wf.total_power = 1
			img_ref = prop(wf)

			wf = Wavefront(aper)
			wf.total_power = 1
			wf.electric_field *= pupil
			wf = vortex(wf)
			wf.electric_field *= lyot
			img = prop(wf)

			assert img.total_power < 1e-6
			assert img.intensity.max() / img_ref.intensity.max() < 1e-8

			pupil, lyot = make_ravc_masks(co, charge)
			pupil = evaluate_supersampled(pupil, pupil_grid, 4)
			lyot = evaluate_supersampled(lyot, pupil_grid, 4)

			transmission = ((pupil * lyot)**2).sum() / (aperture**2).sum()
			transmission_theoretical = get_ravc_planet_transmission(co, charge)

			assert abs(transmission - transmission_theoretical) < 0.01

def test_app_keller():
	pupil_grid = make_pupil_grid(256)
	focal_grid = make_focal_grid(4, 32)
	propagator = FraunhoferPropagator(pupil_grid, focal_grid)

	aperture = evaluate_supersampled(circular_aperture(1), pupil_grid, 8)
	wavefront = Wavefront(aperture)
	wavefront.total_power = 1

	# reference PSF without APP
	img_ref = propagator.forward(wavefront)

	# small rectangular dark zone with 1e-7 contrast
	mask = rectangular_aperture(size=(6,2), center=(9,0))(focal_grid)
	contrast = 1 - mask + 1e-7

	# APP with aggressive acceleration
	app = generate_app_keller(wavefront, propagator, contrast, num_iterations=150, beta=0.98)
	img = propagator.forward(app)

	assert np.abs(img.intensity.max() / img_ref.intensity.max() - 0.947) < 0.01 # Strehl
	assert np.mean(img.intensity * mask) / np.mean(mask) < 1.6e-8 # contrast

def test_perfect_coronagraph():
	pupil_grid = make_pupil_grid(256)
	aperture = circular_aperture(1)(pupil_grid)

	tilts = np.logspace(-3, -1, 51)

	for order in [2, 4, 6, 8]:
		coro = PerfectCoronagraph(aperture, order)

		# Test suppression for on-axis point source
		wf = Wavefront(aperture)
		wf.total_power = 1
		assert coro(wf).total_power < 1e-10

		# Test suppression off-axis
		coronagraph_leakage = []
		for tilt in tilts:
			leakage = coro(Wavefront(aperture * np.exp(2j * np.pi * pupil_grid.x * tilt))).total_power
			coronagraph_leakage.append(leakage)

		y = np.log10(coronagraph_leakage)
		x = np.log10(tilts)
		n = len(x)

		# Do a linear fit on the log-log data to get the power-law coefficient
		beta = ((x * y).sum() - x.sum() * y.sum() / n) / ((x * x).sum() - x.sum()**2 / n)
		assert np.abs(beta - order) / order < 1e-3



import numpy as np

class Field(np.ndarray):
	'''The value of some physical quantity for each point in some coordinate system.

	Parameters
	----------
	arr : array_like
		An array of values or tensors for each point in the :class:`Grid`.
	grid : Grid
		The corresponding :class:`Grid` on which the values are set.

	Attributes
	----------
	grid : Grid
		The grid on which the values are defined.
	'''
	def __new__(cls, arr, grid):
		obj = np.asarray(arr).view(cls)
		obj.grid = grid
		return obj

	def __array_finalize__(self, obj):
		if obj is None:
			return
		self.grid = getattr(obj, 'grid', None)

	@classmethod
	def from_dict(cls, tree):
		'''Make a Field from a dictionary, previously created by `to_dict()`.

		Parameters
		----------
		tree : dictionary
			The dictionary from which to make a new Field object.

		Returns
		-------
		Field
			The created object.

		Raises
		------
		ValueError
			If the dictionary is not formatted correctly.
		'''
		from .grid import Grid

		return cls(np.array(tree['values']), Grid.from_dict(tree['grid']))

	def to_dict(self):
		'''Convert the object to a dictionary for serialization.

		Returns
		-------
		dictionary
			The created dictionary.
		'''
		tree = {
			'values': np.asarray(self),
			'grid': self.grid.to_dict()
		}

		return tree

	def __getstate__(self):
		'''Get the internal state for pickling.

		Returns
		-------
		tuple
			The state of the Field.
		'''
		data_state = super().__reduce__(self)[2]
		return data_state + (self.grid,)

	def __setstate__(self, state):
		'''Set the internal state for pickling.

		Parameters
		----------
		state : tuple
			The state coming from a __getstate__().
		'''
		_, shp, typ, isf, raw, grid = state

		super().__setstate__((shp, typ, isf, raw))
		self.grid = grid

	def __reduce__(self):
		'''Return a 3-tuple for pickling the Field.

		Returns
		-------
		tuple
			The reduced version of the Field.
		'''
		return (_field_reconstruct,
			(self.__class__, np.ndarray, (0,), 'b',),
			self.__getstate__())

	@property
	def tensor_order(self):
		'''The order of the tensor of the field.
		'''
		return self.ndim - 1

	@property
	def tensor_shape(self):
		'''The shape of the tensor of the field.
		'''
		return np.array(self.shape)[:-1]

	@property
	def is_scalar_field(self):
		'''True if this field is a scalar field (ie. a tensor order of 0), False otherwise.
		'''
		return self.tensor_order == 0

	@property
	def is_vector_field(self):
		'''True if this field is a vector field (ie. a tensor order of 1), False otherwise.
		'''
		return self.tensor_order == 1

	@property
	def is_valid_field(self):
		'''True if the field corresponds with its grid.
		'''
		return self.shape[-1] == self.grid.size

	@property
	def shaped(self):
		'''The reshaped version of this field.

		Raises
		------
		ValueError
			If this field isn't separated, no reshaped version can be made.
		'''
		if not self.grid.is_separated:
			raise ValueError('This field doesn\'t have a shape.')

		if self.tensor_order > 0:
			new_shape = np.concatenate([np.array(self.shape)[:-1], self.grid.shape])
			return self.reshape(new_shape)

		return self.reshape(self.grid.shape)

	def at(self, p):
		'''The value of this field closest to point p.

		Parameters
		----------
		p : array_like
			The point at which the closest value should be returned.

		Returns
		-------
		array_like
			The value, potentially tensor, closest to point p.
		'''
		i = self.grid.closest_to(p)
		return self[..., i]

def _field_reconstruct(subtype, baseclass, baseshape, basetype):
	'''Internal function for building a new Field object for pickling.

	Parameters
	----------
	subtype
		The class of Field.
	baseclass
		The array class that was used for the Field.
	baseshape
		The shape of the Field.
	basetype
		The data type of the Field.

	Returns
	-------
	Field
		The built Field object.
	'''
	data = np.ndarray.__new__(baseclass, baseshape, basetype)
	grid = None

	return subtype.__new__(subtype, data, grid)



import numpy as np
from ..mode_basis import ModeBasis
from ..field import Field
from .deformable_mirror import DeformableMirror

class TipTiltMirror(DeformableMirror):
	'''A tip-tilt mirror.

	This deformable mirror class provides a tip-tilt mirror, where the
	actuators are controlled in OPD.

	Parameters
	----------
	input_grid : Grid
		The grid that corresponds to the input wavefront
	'''
	def __init__(self, input_grid):
		self.input_grid = input_grid
		self.actuators = np.zeros((2,))

		modes = ModeBasis([Field(self.input_grid.x, self.input_grid), Field(self.input_grid.y, self.input_grid)])

		super(TipTiltMirror, self).__init__(modes)



import numpy as np
from hcipy import *
import scipy.special
import pytest

def test_fraunhofer_propagation_circular():
	for num_pix in [512, 1024]:
		pupil_grid = make_pupil_grid(num_pix)
		focal_grid = make_focal_grid(16, 8)

		ind = focal_grid.closest_to((0, 0))

		for diameter in [1, 0.7]:
			aperture = evaluate_supersampled(circular_aperture(diameter), pupil_grid, 8)

			for focal_length in [1, 0.8]:
				prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

				for wavelength in [1, 0.5]:
					wf = Wavefront(aperture, wavelength)
					img = prop(wf).electric_field
					img /= img[np.argmax(np.abs(img))]

					x = focal_grid.as_('polar').r * np.pi / wavelength * diameter / focal_length
					x[ind] = 1
					reference = 2 * scipy.special.jv(1, x) / x
					reference[ind] = 1

					if num_pix == 512:
						assert np.abs(img - reference).max() < 3e-5
					elif num_pix == 1024:
						assert np.abs(img - reference).max() < 1e-5
					else:
						assert False # This should never happen.

def test_fraunhofer_propagation_rectangular():
	for num_pix in [512, 1024]:
		pupil_grid = make_pupil_grid(num_pix)
		focal_grid = make_focal_grid(16, 8)

		for size in [[1,1], [0.75,1], [0.75,0.75]]:
			aperture = evaluate_supersampled(rectangular_aperture(size), pupil_grid, 8)

			for focal_length in [1, 1.3]:
				prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

				for wavelength in [1, 0.8]:
					wf = Wavefront(aperture, wavelength)
					img = prop(wf).electric_field
					img /= img[np.argmax(np.abs(img))]

					k_x, k_y = np.array(size) / wavelength / focal_length
					reference = (np.sinc(k_x * focal_grid.x) * np.sinc(k_y * focal_grid.y))

					if num_pix == 512:
						assert np.abs(img - reference).max() < 5e-5
					elif num_pix == 1024:
						assert np.abs(img - reference).max() < 2e-5
					else:
						assert False # This should never happen.

def single_propagation_test(propagator, number_of_pixels, wavelength, a, b, relative_distance, error_threshold):
	wavenumber = 2 * np.pi / wavelength

	pupil_grid = make_pupil_grid(number_of_pixels, [16 * a, 16 * b])
	threshold_distance = np.min(pupil_grid.delta * np.array([16 * a, 16 * b]) / wavelength)
	distance = relative_distance * threshold_distance

	prop_pos = propagator(pupil_grid, distance, num_oversampling=2)
	prop_neg = propagator(pupil_grid, -distance, num_oversampling=2)

	aperture = evaluate_supersampled(rectangular_aperture([2 * a, 2 * b]), pupil_grid, 2)

	img_forward = prop_pos.forward(Wavefront(aperture, wavelength)).intensity
	img_backward = prop_pos.backward(Wavefront(aperture, wavelength)).intensity

	img_forward_neg = prop_neg.forward(Wavefront(aperture, wavelength)).intensity
	img_backward_neg = prop_neg.forward(Wavefront(aperture, wavelength)).intensity

	assert np.allclose(img_forward, img_backward_neg)
	assert np.allclose(img_backward, img_forward_neg)

	def generate_reference_field(reference_grid):
		w_x1 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.x - a)
		w_x2 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.x + a)
		w_y1 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.y - b)
		w_y2 = np.sqrt(2 / (distance * wavelength)) * (reference_grid.y + b)

		fresnel = scipy.special.fresnel

		ssa_x1, csa_x1 = fresnel(w_x1)
		ssa_x2, csa_x2 = fresnel(w_x2)
		ssa_y1, csa_y1 = fresnel(w_y1)
		ssa_y2, csa_y2 = fresnel(w_y2)

		F_x1 = csa_x1 + 1j * ssa_x1
		F_x2 = csa_x2 + 1j * ssa_x2
		F_y1 = csa_y1 + 1j * ssa_y1
		F_y2 = csa_y2 + 1j * ssa_y2

		reference = (F_x2 - F_x1) * (F_y2 - F_y1)
		reference *= -1j * np.exp(1j * wavenumber * distance) / 2
		reference *= np.exp(1j * wavenumber * (reference_grid.x**2 + reference_grid.y**2) / (2 * distance))
		reference = np.abs(Field(reference, reference_grid))**2

		return reference

	reference_image = evaluate_supersampled(generate_reference_field, pupil_grid, 2)

	img_forward = subsample_field(img_forward, 8)
	img_backward = subsample_field(img_backward, 8)
	reference_image = subsample_field(reference_image, 8)

	absolute_error_forward = np.abs(img_forward - reference_image).max()
	assert absolute_error_forward < error_threshold

	absolute_error_backward = np.abs(img_backward - reference_image).max()
	assert absolute_error_backward < error_threshold

def test_fresnel_propagation_rectangular_quick():
	for num_pix in [512,]:
		for wavelength in [500e-9,]:
			for a, b in [[0.001, 0.001],]:
				for relative_distance, error_threshold in zip([0.2, 5],[0.02, 0.01]):
					single_propagation_test(FresnelPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

def test_asp_propagation_rectangular_quick():
	for num_pix in [512,]:
		for wavelength in [500e-9,]:
			for a, b in [[0.001, 0.001],]:
				for relative_distance, error_threshold in zip([0.2, 5],[0.02, 0.01]):
					single_propagation_test(AngularSpectrumPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

@pytest.mark.slow
def test_fresnel_propagation_rectangular():
	for num_pix in [512, 1024]:
		for wavelength in [500e-9, 700e-9]:
			for a, b in [[0.001, 0.001], [0.0015, 0.001]]:
				for relative_distance, error_threshold in zip([0.2, 1, 5],[0.02, 0.01, 0.01]):
					single_propagation_test(FresnelPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)

@pytest.mark.slow
def test_asp_propagation_rectangular():
	for num_pix in [512, 1024]:
		for wavelength in [500e-9, 700e-9]:
			for a, b in [[0.001, 0.001], [0.0015, 0.001]]:
				for relative_distance, error_threshold in zip([0.2, 1, 5],[0.02, 0.01, 0.01]):
					single_propagation_test(AngularSpectrumPropagator, num_pix, wavelength, a, b, relative_distance, error_threshold)



import numpy as np
from .detector import Detector
from .optical_element import OpticalElement, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from .wavefront import Wavefront
from ..mode_basis import ModeBasis, make_LP_modes
from ..field import Field

class StepIndexFiber(AgnosticOpticalElement):
	'''A step-index fiber.

	The modal behaviour of the step-index fiber depends on the input parameters.

	Parameters
	----------
	core_radius : scalar
		The radius of the high-index core.
	NA : scalar or function of wavelength
		The numerical aperture of the fiber.
	fiber_length : scalar
		The length of the optical fiber.
	'''
	def __init__(self, core_radius, NA, fiber_length):
		self._core_radius = core_radius
		self._NA = NA
		self.fiber_length = fiber_length

		AgnosticOpticalElement.__init__(self, False, True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		monochromatic_V = self.V(wavelength)
		instance_data.NA = self.evaluate_parameter(self._NA, input_grid, output_grid, wavelength)
		instance_data.fiber_modes, instance_data.beta = make_LP_modes(input_grid, monochromatic_V, self.core_radius, return_betas=True)

	def num_modes(self, wavelength):
		'''The approximate amount of modes of the fiber.
		'''
		V = self.V(wavelength)
		return V**2 / 2

	def V(self, wavelength):
		'''The normalized frequency parameter for step-index fibers.
		'''
		return 2 * np.pi / wavelength * self.core_radius * self.NA

	def mode_field_radius(self, wavelength):
		'''The mode field radius of the fiber.

		The mode field radius is the radius of the gaussian beam best matched to the fundamental mode [Marcuse1977]_.

		.. [Marcuse1977] D. Marcuse 1977, "Loss analysis of single-mode fiber splices," The Bell System Technical Journal 56, 703-718 (2014)
		'''
		V = self.V(wavelength)
		w = self.core_radius * (0.65 + 1.619 / V**(3 / 2) + 2.879 / V**6)
		return w

	@property
	def core_radius(self):
		'''The core radius of this fiber.
		'''
		return self._core_radius

	@core_radius.setter
	def core_radius(self, core_radius):
		self._core_radius = core_radius
		self.clear_cache()

	@property
	def NA(self):
		'''The numerical aperture of this fiber.
		'''
		return self._NA

	@NA.setter
	def NA(self, NA):
		self._NA = NA
		self.clear_cache()

	def get_input_grid(self, output_grid, wavelength):
		return output_grid

	def get_output_grid(self, input_grid, wavelength):
		return input_grid

	@make_agnostic_forward
	def modal_decomposition(self, instance_data, wavefront):
		'''Decompose the input wavefront into the modal distribution of the fiber.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		array_like
			The modal coefficients.
		'''

		M = instance_data.fiber_modes.transformation_matrix
		mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
		return mode_coefficients

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		'''Forward propagate the light through the fiber.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront that exits the fiber.
		'''

		M = instance_data.fiber_modes.transformation_matrix
		mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
		output_electric_field = np.einsum('...i, ij->...j', mode_coefficients * np.exp(1j * instance_data.beta * self.fiber_length), M.T.conj())
		output_electric_field = Field(output_electric_field, wavefront.grid)

		return Wavefront(output_electric_field, wavefront.wavelength)

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		'''Backward propagate the light through the fiber.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront that exits the fiber.
		'''
		M = instance_data.fiber_modes.transformation_matrix
		mode_coefficients = np.einsum('...i, ij->...j', wavefront.electric_field * wavefront.grid.weights, M.conj())
		output_electric_field = np.einsum('...i, ij->...j', mode_coefficients * np.exp(-1j * instance_data.beta * self.fiber_length), M.T.conj())
		output_electric_field = Field(output_electric_field, wavefront.grid)

		return Wavefront(output_electric_field, wavefront.wavelength)

class SingleModeFiber(Detector):
	def __init__(self, input_grid, mode_field_diameter, mode=None):
		self.input_grid = input_grid
		self.mode_field_diameter = mode_field_diameter

		if mode is None:
			mode = gaussian_mode

		self.mode = mode(self.input_grid, mode_field_diameter)
		self.mode /= np.sum(np.abs(self.mode)**2 * self.input_grid.weights)
		self.intensity = 0

	def integrate(self, wavefront, dt, weight=1):
		self.intensity += weight * dt * (np.dot(wavefront.electric_field * wavefront.electric_field.grid.weights, self.mode))**2

	def read_out(self):
		intensity = self.intensity
		self.intensity = 0
		return intensity

# This implementation assumes orthogonality of the different fibers.
# Forward() should be changed if this is added in the future.
# Also, the modes are independent of wavelength.
class SingleModeFiberArray(OpticalElement):
	def __init__(self, input_grid, fiber_grid, mode, *args, **kwargs):
		'''An array of single-mode fibers.

		Parameters
		----------
		input_grid : Grid
			The grid on which the input wavefront is defined.
		fiber_grid : Grid
			The centers of each of the single-mode fibers.
		mode : function
			The mode of the single-mode fibers. The function should take a grid
			and return the amplitude of the fiber mode.
		'''
		self.input_grid = input_grid
		self.fiber_grid = fiber_grid

		self.fiber_modes = [mode(input_grid.shifted(-p), *args, **kwargs) for p in fiber_grid]
		self.fiber_modes = [m / np.sqrt(np.sum(np.abs(m)**2 * input_grid.weights)) for m in self.fiber_modes]
		self.fiber_modes = ModeBasis(self.fiber_modes)

		self.projection_matrix = self.fiber_modes.transformation_matrix

	def forward(self, wavefront):
		res = self.projection_matrix.T.dot(wavefront.electric_field * self.input_grid.weights)
		return Wavefront(Field(res, self.fiber_grid), wavefront.wavelength)

	def backward(self, wavefront):
		res = self.projection_matrix.dot(wavefront.electric_field)
		return Wavefront(Field(res, self.input_grid), wavefront.wavelength)

	def get_transformation_matrix_forward(self, wavelength=1):
		return self.projection_matrix.T

	def get_transformation_matrix_backward(self, wavelength=1):
		return self.projection_matrix



from hcipy import *
import numpy as np

def check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims):
	grid = make_uniform_grid(dims, 1, has_center=True).shifted(shift_input).scaled(scale)
	f_in = Field(np.random.randn(grid.size), grid).astype(dtype)

	energy_in = np.sum(np.abs(f_in)**2 * f_in.grid.weights)

	fft1 = FastFourierTransform(grid, q=q, fov=fov, shift=shift_output, emulate_fftshifts=True)
	fft2 = FastFourierTransform(grid, q=q, fov=fov, shift=shift_output, emulate_fftshifts=False)
	mft1 = MatrixFourierTransform(grid, fft1.output_grid, precompute_matrices=True, allocate_intermediate=True)
	mft2 = MatrixFourierTransform(grid, fft1.output_grid, precompute_matrices=True, allocate_intermediate=False)
	mft3 = MatrixFourierTransform(grid, fft1.output_grid, precompute_matrices=False, allocate_intermediate=True)
	mft4 = MatrixFourierTransform(grid, fft1.output_grid, precompute_matrices=False, allocate_intermediate=False)
	nft1 = NaiveFourierTransform(grid, fft1.output_grid, precompute_matrices=True)
	nft2 = NaiveFourierTransform(grid, fft1.output_grid, precompute_matrices=False)

	fourier_transforms = [fft1, fft2, mft1, mft2, mft3, mft4, nft1, nft2]

	energy_ratios = []
	patterns_match = []
	for ft1 in fourier_transforms:
		f_inter = ft1.forward(f_in)
		assert f_inter.dtype == np.dtype(dtype)

		for ft2 in fourier_transforms:
			f_out = ft2.backward(f_inter)
			assert f_out.dtype == np.dtype(dtype)

			energy_out = np.sum(np.abs(f_out)**2 * f_out.grid.weights)
			energy_ratio = energy_out / energy_in

			pattern_match = np.std(np.abs(f_out - f_in)) / np.abs(f_in).mean()

			energy_ratios.append(energy_ratio)
			patterns_match.append(pattern_match)

	energy_ratios = np.array(energy_ratios).reshape((len(fourier_transforms), len(fourier_transforms)))
	patterns_match = np.array(patterns_match).reshape((len(fourier_transforms), len(fourier_transforms)))

	if fov == 1:
		# When the full fov is retained, the pattern should be the same and energy should
		# be conserved. We use different accuracy limits based on bit depth.
		#print(np.max(patterns_match))
		if np.dtype(dtype) == np.dtype('complex128'):
			assert np.all(patterns_match < 1e-13)
			assert np.all(np.abs(energy_ratios - 1) < 1e-14)
		else:
			assert np.all(patterns_match < 1e-6)
			assert np.all(np.abs(energy_ratios - 1) < 1e-6)
	else:
		# If the full fov is not retained, the pattern and energy loss should be the same
		# for all fourier transform combinations.
		assert np.allclose(energy_ratios, energy_ratios[0, 0])
		assert np.allclose(patterns_match, patterns_match[0, 0])

def test_fourier_energy_conservation_1d():
	np.random.seed(0)

	for dtype in ['complex128', 'complex64']:
		for shift_input in [0,0.1]:
			for scale in [1,2]:
				for shift_output in [0,0.1]:
					for q in [1,3,4]:
						for fov in [1, 0.5, 0.8]:
							for dims in [64, 65]:
								check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims)

def test_fourier_energy_conservation_2d():
	np.random.seed(0)

	for dtype in ['complex128', 'complex64']:
		for shift_input in [[0,0],[0.1]]:
			for scale in [1,2]:
				for shift_output in [[0,0], [0.1]]:
					for q in [1,3,4]:
						for fov in [1,0.5,0.8]:
							for dims in [[8,8],[8,16],[9,9],[9,18]]:
								check_energy_conservation(dtype, shift_input, scale, shift_output, q, fov, dims)

def check_symmetry(scale, q, fov, dims):
	pass

def test_fourier_symmetries_2d():
	for scale in [1,2]:
		for q in [1,3,4]:
			for fov in [1,0.5,0.8]:
				for dims in [[8,8],[8,16],[9,9],[9,18]]:
					check_symmetry(scale, q, fov, dims)

def test_make_fourier_transform():
	input_grid = make_pupil_grid(128)

	ft = make_fourier_transform(input_grid, q=1, fov=1, planner='estimate')
	assert type(ft) == FastFourierTransform

	ft = make_fourier_transform(input_grid, q=8, fov=0.3, planner='estimate')
	assert type(ft) == MatrixFourierTransform

	ft = make_fourier_transform(input_grid, q=1, fov=1, planner='measure')
	ft = make_fourier_transform(input_grid, q=8, fov=0.1, planner='measure')

	output_grid = CartesianGrid(UnstructuredCoords([np.random.randn(100), np.random.randn(100)]))
	ft = make_fourier_transform(input_grid, output_grid)
	assert type(ft) == NaiveFourierTransform

def test_mft_precomputations():
	input_grid = make_pupil_grid(128)
	output_grid = make_fft_grid(input_grid, 1, 0.25)

	for precompute_matrices in [True, False]:
		for allocate_intermediate in [True, False]:
			mft = MatrixFourierTransform(input_grid, output_grid,
				precompute_matrices=precompute_matrices, allocate_intermediate=allocate_intermediate)

			mft.forward(input_grid.zeros())
			mft.forward(input_grid.ones())

			assert (mft.M1 is not None) == precompute_matrices
			assert (mft.intermediate_array is not None) == allocate_intermediate

def test_fourier_filter():
	for n in [16, 17, [16, 17]]:
		for q in [1, 2, 3]:
			for tensor_shape in [(), (3,), (3,3)]:
				for scale in [1, 2]:
					input_grid = make_pupil_grid(n, scale)

					fft = FastFourierTransform(input_grid, q)

					tf_shape = tensor_shape + (fft.output_grid.size,)
					transfer_function = np.random.randn(*tf_shape) + 1j * np.random.randn(*tf_shape)
					transfer_function = Field(transfer_function, fft.output_grid)

					fourier_filter = FourierFilter(input_grid, transfer_function, q)

					f_shape = tensor_shape + (input_grid.size,)
					if len(tensor_shape) == 2:
						f_shape = f_shape[1:]
					f_in = Field(np.random.randn(*f_shape) + 1j * np.random.randn(*f_shape), input_grid)

					f_out_ff = fourier_filter.forward(f_in)
					f_in_ff = fourier_filter.backward(f_out_ff)

					if len(tensor_shape) == 2:
						ft = fft.forward(f_in)
						ft = field_dot(transfer_function, ft)
						f_out_fft = fft.backward(ft)

						ft = fft.forward(f_out_fft)
						ft = field_dot(field_conjugate_transpose(transfer_function), ft)
						f_in_fft = fft.backward(ft)
					else:
						f_out_fft = fft.backward(fft.forward(f_in) * transfer_function)
						f_in_fft = fft.backward(fft.forward(f_out_fft) * transfer_function.conj())

					assert np.allclose(f_out_fft, f_out_ff)
					assert np.allclose(f_in_fft, f_in_ff)



import numpy as np
import matplotlib as mpl

def errorfill(x, y, y_err_pos, y_err_neg=None, color=None, alpha_fill=0.25, ax=None, **kwargs):
	"""
	Plot a line with filled errorbars.
	"""
	import matplotlib.pyplot as plt

	if ax is None:
		ax = plt.gca()

	if y_err_neg is None:
		y_err_neg = y_err_pos

	y_min = np.array(y) - np.array(y_err_neg)
	y_max = np.array(y) + np.array(y_err_pos)
	if color is None:
		l = ax.plot(x, y, **kwargs)
		color = l[0].get_color()
	else:
		l = ax.plot(x, y, color=color, **kwargs)

	facecolor = mpl.colors.colorConverter.to_rgba(color, alpha=alpha_fill)
	edgecolor = (0,0,0,0)
	ax.fill_between(x, y_max, y_min, edgecolor=edgecolor, facecolor=facecolor)

	return l



import numpy as np
from math import pi

def binned_profile(y, x, bins=20, statistic='mean'):
	'''Create a profile of y(x) in the specified bins.

	Parameters
	----------
	y : array_like
		The y-coordinates of the data points. This should be 1-dimensional.
	x : array_like
		The x-coordinates of the data points. This should be 1-dimensional
		and the same size as `y`.
	bins : array_like or integer
		The bin edges of the profile. If this is an integer, `bins` is the
		number of bins that will be equally distributed along the whole range
		of `x`.
	statistic : string
		The statistic to compute (default is 'mean').
		The following statistics are available:
		* 'mean': compute the mean of values for points within the bin edges.
		* 'sum': compute the sum of values for points within the bin edges.
		* 'min': compute the minimum of values for points within the bin edges.
		* 'max': compute the maximum of values for point within the bin edges.

	Returns
	-------
	bin_centers : array_like
		The center of each of the bins.
	profile : array_like
		The y-values of the resulting profile.
	std_profile : array_like
		The standard deviation within each bin.
	num_per_bin : array_like
		The number of samples per bin.

	Raises
	------
	RuntimeError
		If the number of bins are negative or zero.
	ValueError
		If the statistic is not implemented.
	'''
	if np.isscalar(bins):
		if bins <= 0:
			raise RuntimeError('The number of bins should be positive.')

		# Equally space the bins.
		bins = np.linspace(np.min(x), np.max(x), bins+1)

	bin_centers = (bins[1:] + bins[:-1]) / 2
	num_bins = len(bin_centers)

	num_per_bin = np.histogram(x, bins)[0]
	which_bin = np.digitize(x, bins)

	statistics = {'mean': np.nanmean, 'median': np.nanmedian, 'max': np.nanmax, 'min': np.nanmin}

	if statistic not in statistics:
		raise ValueError('Statistic %s not implemented.' % statistic)

	func = statistics[statistic]

	profile = np.array([func(y[which_bin == b]) for b in range(1, num_bins + 1)])
	std_profile = np.array([np.nanstd(y[which_bin == b]) for b in range(1, num_bins + 1)])

	return bin_centers, profile, std_profile, num_per_bin

def azimutal_profile(image, num_bins, statistic='mean'):
	'''Create an azimuthal profile of the image around its center.

	Parameters
	----------
	image : Field
		The image that we want an azimuthal profile from. This image must be
		two-dimensional.
	num_bins : integer
		The number of bins in theta. Bins will be equally distributed in theta.
	statistic : string
		The statistic to compute (default is 'mean').
		The following statistics are available:
		* 'mean': compute the mean of values for points within the bin edges.
		* 'sum': compute the sum of values for points within the bin edges.
		* 'min': compute the minimum of values for points within the bin edges.
		* 'max': compute the maximum of values for point within the bin edges.

	Returns
	-------
	bin_centers : array_like
		The center of each of the bins.
	profile : array_like
		The y-values of the resulting azimuthal profile.
	std_profile : array_like
		The standard deviation within each bin.
	num_per_bin : array_like
		The number of samples per bin.

	Raises
	------
	RuntimeError
		If the number of bins are negative or zero.
	ValueError
		If the statistic is not implemented.
	'''
	theta = image.grid.as_('polar').theta
	bins = np.linspace(-pi, pi, num_bins + 1)

	return binned_profile(image, theta, bins, statistic=statistic)

def radial_profile(image, bin_size, statistic='mean'):
	'''Create a radial profile of the image around its center.

	Parameters
	----------
	image : Field
		The image that we want an azimuthal profile from. This image must be
		two-dimensional.
	bin_size : scalar
		The extent of each bin. Each bin will be a ring from r to r+`bin_size`.
	statistic : string
		The statistic to compute (default is 'mean').
		The following statistics are available:
		* 'mean': compute the mean of values for points within the bin edges.
		* 'sum': compute the sum of values for points within the bin edges.
		* 'min': compute the minimum of values for points within the bin edges.
		* 'max': compute the maximum of values for point within the bin edges.

	Returns
	-------
	bin_centers : array_like
		The center of each of the bins.
	profile : array_like
		The y-values of the resulting radial profile.
	std_profile : array_like
		The standard deviation within each bin.
	num_per_bin : array_like
		The number of samples per bin.

	Raises
	------
	ValueError
		If the statistic is not implemented.
	'''
	r = image.grid.as_('polar').r

	n_bins = int(np.ceil(r.max() / bin_size))
	max_bin = n_bins * bin_size
	bins = np.linspace(0, max_bin, n_bins + 1)

	return binned_profile(image, r, bins, statistic=statistic)



import numpy as np
from ..field import Field
from .wavefront import Wavefront

class GaussianBeam(object):
	'''An analytical description of a light beam with a Gaussian profile.

	Parameters
	----------
	w0 : scalar
		The beam waist.
	z : scalar
		The current distance from the beam waist.
	wavelength : scalar
		The wavelength of the light.
	'''
	def __init__(self, w0, z, wavelength):
		self.w0 = w0
		self.z = z
		self.wavelength = wavelength

	@property
	def beam_waist(self):
		'''The beam waist of the Gaussian beam.
		'''
		return self.w0

	@beam_waist.setter
	def beam_waist(self, w0):
		self.w0 = w0

	@property
	def zR(self):
		'''The Rayleight distance of the Gaussian beam.
		'''
		return np.pi * self.w0**2 / self.wavelength

	@zR.setter
	def zR(self, zR):
		self.w0 = np.sqrt(zR * self.wavelength / np.pi)

	rayleigh_distance = zR

	@property
	def q(self):
		'''The complex beam parameter of the Gaussian beam.
		'''
		return self.z + 1j * self.zR

	@q.setter
	def q(self, q):
		self.z = np.real(q)
		self.zR = np.imag(q)

	complex_beam_parameter = q

	@property
	def theta(self):
		'''The beam divergence of the Gaussian beam.
		'''
		return self.wavelength / (np.pi * self.w0)

	@theta.setter
	def theta(self, theta):
		self.w0 = self.wavelength / (theta * np.pi)

	beam_divergence = theta

	@property
	def R(self):
		'''The current radius of curvature of the Gaussian beam.
		'''
		epsilon = 1e-16
		if abs(self.z) < epsilon:
			return np.inf
		else:
			return self.z * (1 + (self.zR / self.z)**2)

	radius_of_curvature = R

	@property
	def psi(self):
		'''The current Gouy phase of the Gaussian beam.
		'''
		return np.arctan(self.z / self.zR)

	gouy_phase = psi

	@property
	def w(self):
		'''The current beam radius of the Gaussian beam.
		'''
		return self.w0 * np.sqrt(1 + (self.z / self.zR)**2)

	beam_radius = w

	@property
	def FWHM(self):
		'''The current FWHM of the Gaussian beam.
		'''
		return self.w * np.sqrt(2 * np.log(2))

	full_width_half_maximum = FWHM

	@property
	def k(self):
		'''The wavenumber of the Gaussian beam.
		'''
		return 2 * np.pi / self.wavelength

	@k.setter
	def k(self, k):
		self.wavelength = 2 * np.pi / k

	wavenumber = k

	def evaluate(self, grid):
		'''Evaluate the wavefront of the Gaussian beam at the current position on
		the given grid.

		Parameters
		----------
		grid : Grid
			The grid on which to calculate the wavefront for the Gaussian beam.

		Returns
		-------
		Wavefront
			The evaluated wavefront of the Gaussian beam.
		'''
		if grid.is_('cartesian'):
			r2 = grid.x**2 + grid.y**2
		else:
			r2 = grid.as_('polar').r**2

		K1 = self.w0 / self.w
		K2 = np.exp(-r2 / self.w**2)
		K3 = np.exp(-1j * (self.k * self.z + self.k * r2 / (2 * self.R) - self.psi))

		return Wavefront(Field(K1 * K2 * K3, grid), self.wavelength)

	__call__ = evaluate



import numpy as np
from .optical_element import OpticalElement
from ..field import CartesianGrid, UnstructuredCoords

class PeriodicOpticalElement(OpticalElement):
	def __init__(self, input_grid, pitch, apodization, orientation=0, even_grid=False):
		'''An even asphere micro-lens array.

		Parameters
		----------
		input_grid : Grid
			The grid on which the periodic optical element is evaluated.
		pitch : scalar
			The pitch of the periodic optical element.
		apodization : Apodizer
			The apodizer that will be evaluated on the periodic grid.
		orientation : scalar
			The orientation of the periodic optical element.
		even_grid : bool
			This determines whether zero is in between two elements or if it is the center of an element.
		'''
		self.input_grid = input_grid.copy()
		self.input_grid = self.input_grid.rotated(orientation)

		if even_grid:
			xf = (np.fmod(abs(self.input_grid.x), pitch) - pitch / 2) * np.sign(self.input_grid.x)
			yf = (np.fmod(abs(self.input_grid.y), pitch) - pitch / 2) * np.sign(self.input_grid.y)
		else:
			xf = (np.fmod(abs(self.input_grid.x) + pitch / 2, pitch) - pitch / 2) * np.sign(self.input_grid.x)
			yf = (np.fmod(abs(self.input_grid.y) + pitch / 2, pitch) - pitch / 2) * np.sign(self.input_grid.y)

		periodic_grid = CartesianGrid(UnstructuredCoords((xf, yf)))
		self.apodization = apodization(periodic_grid)

	def forward(self, wavefront):
		wf = wavefront.copy()
		return self.apodization.forward(wf)

	def backward(self, wavefront):
		wf = wavefront.copy()
		return self.apodization.backward(wf)



from hcipy import *
import numpy as np
import os
import warnings
import itertools
import pytest
import functools

def check_against_reference(field_generator, diameter, baseline_name):
	fname = os.path.join(os.path.dirname(__file__), 'baseline_for_apertures/' + baseline_name + '.fits.gz')

	grid = make_uniform_grid(256, [diameter, diameter])
	field = evaluate_supersampled(field_generator, grid, 8)

	if os.path.isfile(fname):
		reference = read_fits(fname).ravel()

		assert np.allclose(field, reference)
	else:
		warnings.warn('Baseline aperture %s not available. Writing reference now...' % fname)

		if not os.path.exists(os.path.dirname(fname)):
			os.makedirs(os.path.dirname(fname))
		write_fits(field, fname)

def check_segmentation(aperture_function, segments=None):
	grid = make_uniform_grid(256, [1, 1])

	if segments is None:
		aperture, segments = aperture_function(normalized=True, return_segments=True)
	else:
		aperture = aperture_function

	aperture = evaluate_supersampled(aperture, grid, 2)
	segments = evaluate_supersampled(segments, grid, 2)

	aperture_from_segments = segments.linear_combination(np.ones(len(segments)))

	assert np.allclose(aperture, aperture_from_segments)

def check_aperture_against_reference(aperture_function, basename, diameter, options, segmented=False):
	keys = sorted(options.keys())
	vals = [value for key, value in sorted(options.items())]

	for option in itertools.product(*vals):
		fname = basename + '/pupil' + ''.join([o[1] for o in option])

		kwargs = dict(zip(keys, [o[0] for o in option]))
		aperture = aperture_function(**kwargs)

		check_against_reference(aperture, 1 if kwargs.get('normalized', False) else diameter, fname)

		if segmented:
			check_segmentation(*aperture_function(return_segments=True, **kwargs))

def test_regular_polygon_aperture():
	options = {
		'num_sides': [(6, '_hex'), (5, '_pent')],
		'angle': [(0, ''), (15, '_rotated')]
	}

	check_aperture_against_reference(functools.partial(regular_polygon_aperture, circum_diameter=1), 'polygon', 1, options)

def test_elliptical_aperture():
	options = {
		'diameters': [([0.5, 0.5], '_round'), ([1, 0.5], '_elongated')]
	}

	check_aperture_against_reference(elliptical_aperture, 'ellipse', 1, options)

def test_obstructed_circular_aperture():
	options = {
		'central_obscuration_ratio': [(0.1, '_small_obscuration'), (0.3, '_large_obscuration')],
		'num_spiders': [(3, '_3spiders'), (5, '_5spiders')],
		'spider_width': [(0.01, '_thinspiders'), (0.03, '_thickspiders')]
	}

	check_aperture_against_reference(functools.partial(make_obstructed_circular_aperture, pupil_diameter=1), 'obstructed_circular', 1, options)

def test_hexagonal_segmented_aperture():
	options = {
		'num_rings': [(3, '_3rings'), (5, '_5rings')],
		'segment_flat_to_flat': [(0.04, '_smallsegment'), (0.07, '_largesegment')],
		'gap_size': [(0.01, '_smallgap'), (0.02, '_largegap')],
		'starting_ring': [(0, '_withcenter'),  (2, '_withoutcenter')]
	}

	check_aperture_against_reference(make_hexagonal_segmented_aperture, 'hexagonal_segmented', 1, options, segmented=True)

def test_vlt_aperture():
	options = {
		'telescope': [('ut1', '_ut123'), ('ut2', '_ut123'), ('ut3', '_ut123'), ('antu', '_ut123'), ('kueyen', '_ut123'), ('melipal', '_ut123')],
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')]
	}

	check_aperture_against_reference(make_vlt_aperture, 'vlt', 8.1196, options)

	options = {
		'telescope': [('ut4', '_ut4'), ('yepun', '_ut4')],
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')],
		'with_M3_cover': [(False, ''), (True, '_with_M3_cover')]
	}

	check_aperture_against_reference(make_vlt_aperture, 'vlt', 8.1196, options)

	for telescope in ['ut1', 'ut2', 'ut3']:
		with pytest.warns(UserWarning, match='Using the M3 cover on a telescope other than UT4 is not realistic.'):
			aper = make_vlt_aperture(telescope=telescope, with_M3_cover=True)

	with pytest.raises(ValueError):
		aper = make_vlt_aperture(telescope='nonexistent_vlt_telescope')

def test_magellan_aperture():
	options = {
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')]
	}

	check_aperture_against_reference(make_magellan_aperture, 'magellan', 6.5, options)

def test_luvoir_a_aperture():
	check_against_reference(make_luvoir_a_aperture(), 15.0, 'luvoir_a/pupil')

	check_segmentation(make_luvoir_a_aperture)

@pytest.mark.slow
def test_luvoir_a_aperture_all():
	options = {
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')],
		'with_segment_gaps': [(True, ''), (False, '_without_segment_gaps')]
	}

	check_aperture_against_reference(make_luvoir_a_aperture, 'luvoir_a', 15, options)

def test_luvoir_a_lyot_stop():
	options = {
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')]
	}

	check_aperture_against_reference(make_luvoir_a_lyot_stop, 'luvoir_a_lyot', 15, options)

def test_hicat_aperture():
	check_against_reference(make_hicat_aperture(), 0.019725, 'hicat_pupil/pupil')

	check_segmentation(make_hicat_aperture)

@pytest.mark.slow
def test_hicat_aperture_all():
	options = {
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')],
		'with_segment_gaps': [(True, ''), (False, '_without_segment_gaps')]
	}

	check_aperture_against_reference(make_hicat_aperture, 'hicat_pupil', 0.019725, options)

def test_hicat_lyot_stop():
	options = {
		'normalized': [(False, ''), (True, '_normalized')],
		'with_spiders': [(True, ''), (False, '_without_spiders')]
	}

	check_aperture_against_reference(make_hicat_lyot_stop, 'hicat_lyot', 19.9e-3, options)
